"""Unique key creating index."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class UniqueKeyConstraintCreatingIndex(linter.BaseChecker):
    """## **What it does**
    Checks unique key constraint creating index.

    ## **Why not?**
    Unique key constraint is supported by a unique index in the background.
    Adding a unique key constraint to an already populated table will create a
    unique index in non-concurrent mode, scanning and validating that there are no
    violating records in the table. This causes the table to be locked, in which no other
    operations can be performed on the table for the duration of the validation.
    This will cause downtime if the table is concurrently being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    1. Create a unique index in concurrent mode:
        **CREATE UNIQUE INDEX CONCURRENTLY .. ON .. (..)**.
    2. Add the unique constraint using the index created in step 1:
        **ALTER TABLE .. ADD CONSTRAINT .. UNIQUE USING INDEX ..**
    """

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> None:
        """Visit Constraint."""
        if (
            ancestors.find_nearest(ast.AlterTableCmd)
            and node.contype == enums.ConstrType.CONSTR_UNIQUE
            and not node.indexname
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Unique key constraint creating index",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Create a unique index in concurrent mode, then add the unique key constraint using the index",  # noqa: E501
                ),
            )



"""Checker for money."""

from pglast import ast, visitors

from pgrubic.core import linter


class Money(linter.BaseChecker):
    """## **What it does**
    Checks for usage of money.

    ## **Why not?**
    It's a fixed-point type, implemented as a machine int, so arithmetic with it is fast.
    But it doesn't handle fractions of a cent (or equivalents in other currencies), it's
    rounding behaviour is probably not what you want.

    It doesn't store a currency with the value, rather assuming that all money columns
    contain the currency specified by the database's lc_monetary locale setting. If you
    change the lc_monetary setting for any reason, all money columns will contain the
    wrong value. That means that if you insert '$10.00' while lc_monetary is set to
    'en_US.UTF-8' the value you retrieve may be '10,00 Lei' or '¥1,000' if lc_monetary
    is changed.

    Storing a value as a numeric, possibly with the currency being used in an adjacent
    column, might be better.

    ## **When should you?**
    If you're only working in a single currency, aren't dealing with fractional cents
    and are only doing addition and subtraction then money might be the right thing.

    ## **Use instead:**
    numeric
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if node.typeName.names[-1].sval == "money":
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer numeric to money",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use numeric",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.typeName = ast.TypeName(
            names=(
                {
                    "@": "String",
                    "sval": "numeric",
                },
            ),
        )



"""Formatter for boolean expressions."""

from pglast import ast, enums, stream, printers


def bool_expr_needs_to_be_wrapped_in_parens(node: ast.BoolExpr) -> bool:
    """Check if the node needs to be wrapped in parens."""
    return isinstance(node, ast.BoolExpr) and node.boolop in (
        enums.BoolExprType.AND_EXPR,
        enums.BoolExprType.OR_EXPR,
    )


@printers.node_printer(ast.BoolExpr, override=True)
def bool_expr(node: ast.BoolExpr, output: stream.RawStream) -> None:
    """Printer for BoolExpr."""
    in_target_list = isinstance(node.ancestors[0], ast.ResTarget)
    bool_expr_in_ancestors = ast.BoolExpr in node.ancestors

    if node.boolop == enums.BoolExprType.AND_EXPR:
        indent_value = -4 if not in_target_list else None
        relative_indent = (
            -5 if bool_expr_in_ancestors and not in_target_list else indent_value
        )
        output.print_list(
            node.args,
            "AND",
            relative_indent=relative_indent,
            item_needs_parens=bool_expr_needs_to_be_wrapped_in_parens,
        )
    elif node.boolop == enums.BoolExprType.OR_EXPR:
        relative_indent = -3 if not in_target_list else None
        output.print_list(
            node.args,
            "OR",
            relative_indent=relative_indent,
            item_needs_parens=bool_expr_needs_to_be_wrapped_in_parens,
        )
    else:
        output.writes("NOT")
        with output.expression(
            bool_expr_needs_to_be_wrapped_in_parens(
                node.args[0],
            ),
        ):
            output.print_node(node.args[0])



"""Checker for non concurrent detach partition."""

from pglast import ast, visitors

from pgrubic.core import linter


class NonConcurrentDetachPartition(linter.BaseChecker):
    """## **What it does**
    Checks non-concurrent detach partition.

    ## **Why not?**
    Detaching a partition in non-concurrent mode acquires an **ACCESS EXCLUSIVE** lock on
    both the partition and the parent table, blocking other accesses to all partitions
    until the **DETACH PARTITION** is completed.
    This will cause downtime if the partitions are concurrently being accessed by other
    clients.

    Concurrent mode, **CONCURRENTLY** is a new feature from PostgreSQL 14.

    ## **When should you?**
    If the partitioned table is empty.
    If the partitioned table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    From PostgreSQL 14, detach the partition in concurrent mode:
        **ALTER TABLE .. DETACH PARTITION .. CONCURRENTLY**
    """

    is_auto_fixable: bool = True

    def visit_PartitionCmd(
        self,
        ancestors: visitors.Ancestor,
        node: ast.PartitionCmd,
    ) -> None:
        """Visit PartitionCmd."""
        detach_partition_concurrently_postgres_version = 14
        if (
            self.config.lint.postgres_target_version
            >= detach_partition_concurrently_postgres_version
            and not node.concurrent
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Non concurrent detach partition",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Detach the partition in concurrent mode",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.PartitionCmd) -> None:
        """Fix violation."""
        node.concurrent = True



"""Checker for disallowed data types."""

from pglast import ast, visitors

from pgrubic.core import config, linter


class DisallowedDataType(linter.BaseChecker):
    """## **What it does**
    Checks for usage of disallowed data types.

    ## **Why not?**
    If you are using a disallowed type, you're probably doing something wrong.

    ## **When should you?**
    Never. If a data type is intended to be used, it should not be in the
    config **disallowed_data_types**.

    ## **Use instead:**
    Allowed data types.

    ## **Configuration**
    `disallowed_data_types`: List of disallowed data types.
    """

    is_auto_fixable: bool = True

    def visit_TypeName(
        self,
        ancestors: visitors.Ancestor,
        node: ast.TypeName,
    ) -> None:
        """Visit TypeName."""
        for data_type in self.config.lint.disallowed_data_types:
            if node.names[-1].sval == data_type.name:
                self.violations.add(
                    linter.Violation(
                        rule_code=self.code,
                        rule_name=self.name,
                        rule_category=self.category,
                        line_number=self.line_number,
                        column_offset=self.column_offset,
                        line=self.line,
                        statement_location=self.statement_location,
                        description=f"Data type '{node.names[-1].sval}' is disallowed"
                        f" in config with reason: '{data_type.reason}', use"
                        f" '{data_type.use_instead}' instead",
                        is_auto_fixable=self.is_auto_fixable,
                        is_fix_enabled=self.is_fix_enabled,
                        help=f"Use '{data_type.use_instead}' instead",
                    ),
                )

                self._fix(node, data_type)

    def _fix(self, node: ast.TypeName, data_type: config.DisallowedDataType) -> None:
        """Fix violation."""
        node.names = (
            {
                "@": "String",
                "sval": data_type.use_instead,
            },
        )



"""Formatter for SELECT statements."""

from pglast import ast, enums, stream, printers


@printers.node_printer(ast.SubLink, override=True)
def sub_link(node: ast.SubLink, output: stream.RawStream) -> None:
    """Printer for SubLink."""
    if node.subLinkType == enums.SubLinkType.EXISTS_SUBLINK:
        output.write("EXISTS")
        output.space()
    elif node.subLinkType == enums.SubLinkType.ALL_SUBLINK:
        output.print_node(node.testexpr)
        output.space()
        output.write(printers.dml.get_string_value(node.operName))
        output.space()
        output.write("ALL")
        output.space()
    elif node.subLinkType == enums.SubLinkType.ANY_SUBLINK:
        output.print_node(node.testexpr)

        if node.operName:
            output.space()
            output.write(printers.dml.get_string_value(node.operName))
            output.space()
            output.write("ANY")
            output.space()
        else:
            output.space()
            output.write("IN")
            output.space()

    elif node.subLinkType == enums.SubLinkType.ARRAY_SUBLINK:
        output.write("ARRAY")
        output.space()
    elif node.subLinkType in (  # pragma: no cover
        enums.SubLinkType.MULTIEXPR_SUBLINK,
        enums.SubLinkType.ROWCOMPARE_SUBLINK,
    ):
        msg = f"SubLink of type {node.subLinkType} not supported yet"
        raise NotImplementedError(msg)

    with output.expression(need_parens=False):
        bool_in_ancestors = ast.BoolExpr in node.ancestors
        nearest_node: ast.Node = abs(node.ancestors).node

        if (
            isinstance(nearest_node, ast.SelectStmt | ast.UpdateStmt | ast.DeleteStmt)
            and nearest_node.whereClause
        ) or isinstance(nearest_node, ast.A_Expr):
            indent = 11
            dedent = -4
        elif bool_in_ancestors:
            indent = 8
            dedent = -4
        else:
            indent = 6
            dedent = -4

        with output.push_indent(indent, relative=False):
            output.write("(")
            output.newline()
            output.print_node(node.subselect)
            output.newline()
            output.indent(dedent, relative=False)
            output.write(")")
            output.dedent()


def subexpression_needs_parentheses(node: ast.SelectStmt) -> bool:
    """Check if subexpression needs to be wrapped in parentheses in set operations.
    A SELECT statement on either sides of UNION/INTERSECT/EXCEPT must be wrapped in
    parentheses if it contains ORDER BY/LIMIT/... or is a nested UNION/INTERSECT/EXCEPT.
    """
    return bool(
        node.sortClause
        or node.limitCount
        or node.limitOffset
        or node.lockingClause
        or node.withClause
        or node.op != enums.SetOperation.SETOP_NONE,
    )


@printers.node_printer(ast.SelectStmt, override=True)
def select_stmt(node: ast.SelectStmt, output: stream.RawStream) -> None:
    """Printer for SelectStmt."""
    with output.push_indent():
        if node.withClause:
            output.write("WITH")
            output.space()
            output.print_node(node.withClause)
            output.indent()

        if node.valuesLists:
            with output.expression(need_parens=False):
                output.write("VALUES")
                output.space()
                output.print_lists(node.valuesLists, standalone_items=False)
        elif node.op != enums.SetOperation.SETOP_NONE and (node.larg or node.rarg):
            with output.push_indent():
                if node.larg:
                    with output.expression(
                        subexpression_needs_parentheses(
                            node.larg,
                        ),
                    ):
                        output.print_node(node.larg)

                output.newline()

                if node.op == enums.SetOperation.SETOP_UNION:
                    output.space()
                    if not subexpression_needs_parentheses(node.larg):
                        output.write("UNION")
                    else:
                        output.space()
                        output.write("UNION")
                elif node.op == enums.SetOperation.SETOP_INTERSECT:
                    output.write("INTERSECT")
                elif node.op == enums.SetOperation.SETOP_EXCEPT:
                    output.write("EXCEPT")

                if node.all:
                    output.space()
                    output.write("ALL")

                output.newline()

                if node.rarg:
                    with output.expression(
                        subexpression_needs_parentheses(
                            node.rarg,
                        ),
                    ):
                        output.print_node(node.rarg)
        else:
            output.write("SELECT")
            if node.distinctClause:
                output.space()
                output.write("DISTINCT")

                if node.distinctClause[0]:
                    output.space()
                    output.write("ON")
                    output.space()
                    with output.expression(need_parens=True):
                        output.print_list(node.distinctClause)

                output.newline()
                output.space(6)

            if node.targetList:
                output.space()
                output.print_list(node.targetList)

            if node.intoClause:
                output.newline()
                output.space(2)
                output.write("INTO")
                output.space()

                if node.intoClause.rel.relpersistence == enums.RELPERSISTENCE_UNLOGGED:
                    output.write("UNLOGGED")
                    output.space()
                elif node.intoClause.rel.relpersistence == enums.RELPERSISTENCE_TEMP:
                    output.write("TEMPORARY")
                    output.space()

                output.print_node(node.intoClause)

            if node.fromClause:
                output.newline()
                output.space(2)
                output.write("FROM")
                output.space()
                output.print_list(node.fromClause, standalone_items=True)

            if node.whereClause:
                output.newline()
                output.space()
                output.write("WHERE")
                output.space()
                output.print_node(node.whereClause)

            if node.groupClause:
                output.newline()
                output.space()
                output.write("GROUP BY")
                output.space()

                if node.groupDistinct:
                    output.write("DISTINCT")
                    output.space()

                output.print_list(node.groupClause, standalone_items=True)

            if node.havingClause:
                output.newline()
                output.write("HAVING")
                output.space()
                output.print_node(node.havingClause)

            if node.windowClause:
                output.newline()
                output.write("WINDOW")
                output.space()
                output.print_list(node.windowClause)

        if node.sortClause:
            output.newline()
            output.space()
            output.write("ORDER BY")
            output.space()
            output.print_list(node.sortClause)

        if node.limitCount:
            output.newline()

            if node.limitOption == enums.LimitOption.LIMIT_OPTION_COUNT:
                output.space()
                output.write("LIMIT")
            elif node.limitOption == enums.LimitOption.LIMIT_OPTION_WITH_TIES:
                output.space()
                output.write("FETCH FIRST")

            output.space()

            if isinstance(node.limitCount, ast.A_Const) and node.limitCount.isnull:
                output.write("ALL")
            else:
                with output.expression(
                    isinstance(node.limitCount, ast.A_Expr)
                    and node.limitCount.kind == enums.A_Expr_Kind.AEXPR_OP,
                ):
                    output.print_node(node.limitCount)

            if node.limitOption == enums.LimitOption.LIMIT_OPTION_WITH_TIES:
                output.space()
                output.write("ROWS WITH TIES")

        if node.limitOffset:
            output.newline()
            output.write("OFFSET")
            output.space()
            output.print_node(node.limitOffset)

        if node.lockingClause:
            output.newline()
            output.write("FOR")
            output.space()
            output.print_list(node.lockingClause)

        if node.withClause:
            output.dedent()


@printers.node_printer(ast.RangeSubselect, override=True)
def range_subselect(node: ast.RangeSubselect, output: stream.RawStream) -> None:
    """Printer for RangeSubselect."""
    if node.lateral:
        output.write("LATERAL")
        output.space()

    with output.push_indent():
        output.write("(")
        output.newline()
        output.space(4)
        output.print_node(node.subquery)
        output.newline()
        output.write(")")

    if node.alias:
        output.space()
        output.write("AS")
        output.space()
        output.print_name(node.alias)



"""Checker for xml."""

from pglast import ast, visitors

from pgrubic.core import linter


class Xml(linter.BaseChecker):
    """## **What it does**
    Checks for usage of xml.

    ## **Why not?**
    Downsides to using XML include slower processing and querying due to its complexity,
    higher storage requirements and challenges in efficient indexing, rigidity in
    adapting to schema changes, and the overall complexity in data handling due to XML's
    verbose and hierarchical structure. These factors suggest that XML can lead to
    increased resource usage and development difficulties, making formats like JSON more
    suitable.

    ## **When should you?**
    Despite its drawbacks, XML might be suitable when there is a need for data
    interchange with systems that require XML format, or when dealing with legacy
    systems where data is already in XML format.

    ## **Use instead:**
    jsonb.
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if node.typeName.names[-1].sval == "xml":
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer jsonb over xml",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use jsonb",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.typeName = ast.TypeName(
            names=(
                {
                    "@": "String",
                    "sval": "jsonb",
                },
            ),
        )



"""Checker for missing primary key."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class MissingPrimaryKey(linter.BaseChecker):
    """## **What it does**
    Checks for existence of primary keys.

    ## **Why not?**
    A primary key is a set of attributes which prevents duplicates and can uniquely
    identify a record.

    If you don't have one, you cannot tell records apart and you
    create a burden for yourself by checking for duplicate records.

    ## **When should you?**
    Never for OLTP databases.

    If you are doing analytics (OLAP) at scale, then primary
    keys may not be very useful.

    ## **Use instead:**
    Define a primary key.
    """

    def _check_for_table_level_primary_key(
        self,
        node: ast.CreateStmt,
    ) -> bool:
        """Check for table level primary key."""
        return bool(
            (
                [
                    definition
                    for definition in node.tableElts
                    if isinstance(definition, ast.Constraint)
                    and definition.contype == enums.ConstrType.CONSTR_PRIMARY
                ]
            ),
        )

    def _check_for_column_level_primary_key(
        self,
        node: ast.CreateStmt,
    ) -> bool:
        """Check for column level primary key."""
        return bool(
            (
                [
                    definition
                    for definition in node.tableElts
                    if isinstance(definition, ast.ColumnDef) and definition.constraints
                    for constraint in definition.constraints
                    if constraint.contype == enums.ConstrType.CONSTR_PRIMARY
                ]
            ),
        )

    def visit_CreateStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateStmt,
    ) -> None:
        """Visit CreateStmt."""
        if (
            node.tableElts
            and not self._check_for_column_level_primary_key(node)
            and not self._check_for_table_level_primary_key(node)
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Table `{node.relation.relname}` missing a primary key",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Define a primary key",
                ),
            )



"""Checker for identity generated by default."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class IdentityGeneratedByDefault(linter.BaseChecker):
    """## **What it does**
    Checks for identity generated by default.

    ## **Why not?**
    **GENERATED BY DEFAULT** allows you to specify a value for insert or update on an
    identity column instead of system generated value. This is usually not what you
    want for an identity column.

    ## **When should you?**
    If you want to be able to specify a value for insert or update on an identity
    column instead of system generated value.

    ## **Use instead:**
    **GENERATED ALWAYS**
    """

    is_auto_fixable: bool = True

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> None:
        """Visit Constraint."""
        if (
            node.contype == enums.ConstrType.CONSTR_IDENTITY
            and node.generated_when == enums.ATTRIBUTE_IDENTITY_BY_DEFAULT
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer generated always over generated by default identity",  # noqa: E501
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use generated always",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.Constraint) -> None:
        """Fix violation."""
        node.generated_when = enums.ATTRIBUTE_IDENTITY_ALWAYS



"""Checker for cascade delete."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class CascadeDelete(linter.BaseChecker):
    """## **What it does**
    Checks for usage of cascade update.

    ## **Why not?**
    Database schema should follow the principle of least surprise which
    says that every component in a system should behave in a way that most users
    expect it to behave, and therefore not surprise or astonish them.

    Cascading deletes should not cause unexpected loss of data. It is certainly
    dangerous if deleting a single table row can wipe out half your database.

    Are you certain you want to take the chance that someone will delete a single
    reference entry, unaware that doing so may delete a billion related transactions?

    ## **When should you?**
    Almost never.

    ## **Use instead:**
    Restrict
    """

    is_auto_fixable: bool = True

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> None:
        """Visit Constraint."""
        if (
            node.contype == enums.ConstrType.CONSTR_FOREIGN
            and node.fk_del_action == enums.FKCONSTR_ACTION_CASCADE
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Cascade delete in foreign key constraint",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Handle the deletes manually",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.Constraint) -> None:
        """Fix violation."""
        node.fk_del_action = enums.FKCONSTR_ACTION_RESTRICT



"""Checker for table inheritance."""

from pglast import ast, visitors

from pgrubic.core import linter


class TableInheritance(linter.BaseChecker):
    """## **What it does**
    Checks for usage of table inheritance.

    ## **Why not?**
    Table inheritance was a part of a fad wherein the database was closely coupled to
    object-oriented code. It turned out that coupling things that closely didn't
    actually produce the desired results.

    ## **When should you?**
    Never …almost. Now that table partitioning is done natively, that common use case
    for table inheritance has been replaced by a native feature that handles tuple
    routing, etc., without bespoke code. One of the very few exceptions would be
    temporal_tables extension if you are in a pinch and want to use that for row
    versioning in place of a lacking SQL 2011 support. Table inheritance will provide a
    small shortcut instead of using UNION ALL to get both historical as well as current
    rows. Even then you ought to be wary of caveats while working with parent table.

    ## **Use instead:**
    Don't use table inheritance. Use declarative partitioning.
    """

    def visit_CreateStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateStmt,
    ) -> None:
        """Visit CreateStmt."""
        if node.inhRelations and not node.partbound:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Table inheritance detected",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use declarative partitioning",
                ),
            )



"""Checker for mismatch column in data type change."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class MismatchColumnInDataTypeChange(linter.BaseChecker):
    """## **What it does**
    Checks for mismatch column in data type change.

    ## **Why not?**
    For certain column data type changes, a **USING** clause must be provided if there is
    no implicit or assignment cast from old to new type.
    Logically, the expression in the USING should reference the original column otherwise
    it is most likely a mistake.

    ## **When should you?**
    Almost never. When you are sure that the expression in the USING is indeed correct.

    ## **Use instead:**
    The right column in the USING clause.
    """

    def visit_ColumnRef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnRef,
    ) -> None:
        """Visit ColumnRef."""
        alter_table_cmd: visitors.Ancestor = ancestors.find_nearest(ast.AlterTableCmd)

        if (
            alter_table_cmd
            and alter_table_cmd.node.subtype == enums.AlterTableType.AT_AlterColumnType
            and alter_table_cmd.node.name != node.fields[-1].sval
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Column `{alter_table_cmd.node.name}` in data type"
                    f" change does not match column `{node.fields[-1].sval}`"
                    " in USING clause",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use the right column in the USING clause",
                ),
            )



"""Checker for smallint."""

from pglast import ast, visitors

from pgrubic.core import linter


class Smallint(linter.BaseChecker):
    """## **What it does**
    Checks for usage of smallint.

    ## **Why not?**
    Smallint can store values up to 32767 which can lead to integer overflow once
    the max value is reached. The fire drill when you run out of integers is not cheap.

    ## **When should you?**
    Tables that store limited lookup options.

    ## **Use instead:**
    bigint.
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if node.typeName.names[-1].sval == "int2":
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer bigint over smallint",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use bigint",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.typeName = ast.TypeName(
            names=(
                {
                    "@": "String",
                    "sval": "bigint",
                },
            ),
        )



"""Entry point."""

import os
import sys
import typing
import difflib
import logging
import pathlib
import multiprocessing
from collections import abc

import click
from rich.syntax import Syntax
from rich.console import Console

from pgrubic import PACKAGE_NAME, DEFAULT_WORKERS, WORKERS_ENVIRONMENT_VARIABLE, core
from pgrubic.core import noqa, errors

T = typing.TypeVar("T")


def common_options(func: abc.Callable[..., T]) -> abc.Callable[..., T]:
    """Decorator to add common options to each subcommand."""
    func = click.version_option()(func)
    func = click.option("--workers", type=int, help="Number of workers to use.")(func)
    return click.option("--verbose", is_flag=True, help="Enable verbose logging.")(func)


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    epilog=f"""
Examples:\n
   {PACKAGE_NAME} lint\n
   {PACKAGE_NAME} lint .\n
   {PACKAGE_NAME} lint *.sql\n
   {PACKAGE_NAME} lint example.sql\n
   {PACKAGE_NAME} format file.sql\n
   {PACKAGE_NAME} format migrations/\n
""",
)
@click.version_option()
def cli() -> None:
    """Pgrubic: PostgreSQL linter and formatter for schema migrations
    and design best practices.
    """


@cli.command(name="lint")
@click.option(
    "--fix",
    is_flag=True,
    default=False,
    help="Fix lint violations automatically.",
)
@click.option(
    "--ignore-noqa",
    is_flag=True,
    default=False,
    help="Whether to ignore noqa directives.",
)
@click.option(
    "--add-file-level-general-noqa",
    is_flag=True,
    default=False,
    help="Whether to add file-level noqa directives.",
)
@common_options
@click.argument("sources", nargs=-1, type=click.Path(exists=True, path_type=pathlib.Path))  # type: ignore [type-var]
def lint(  # noqa: C901, PLR0912, PLR0913
    sources: tuple[pathlib.Path, ...],
    *,
    fix: bool,
    ignore_noqa: bool,
    add_file_level_general_noqa: bool,
    workers: int,
    verbose: bool,
) -> None:
    """Lint SQL files.

    Parameters:
    ----------
    sources: tuple[pathlib.Path, ...]
        List of sources to lint.
    fix: bool
        Fix lint violations automatically.
    ignore_noqa: bool
        Whether to ignore noqa directives.
    add_file_level_general_noqa: bool
        Whether to add file-level noqa directives.
    workers: int
        Number of workers to use.
    verbose: bool
        Enable verbose logging.

    Returns:
    -------
    None

    """
    core.logger.setLevel(logging.INFO if verbose else logging.WARNING)

    config: core.Config = core.parse_config()

    for key, value in [("fix", fix), ("ignore_noqa", ignore_noqa)]:
        if value:
            setattr(config.lint, key, value)

    linter: core.Linter = core.Linter(config=config, formatters=core.load_formatters)

    rules: set[core.BaseChecker] = core.load_rules(config=config)

    for rule in rules:
        linter.checkers.add(rule())

    total_violations = 0
    auto_fixable_violations = 0
    fix_enabled_violations = 0

    # Use the current working directory if no sources are specified
    if not sources:
        sources = (pathlib.Path.cwd(),)

    included_sources: set[pathlib.Path] = core.filter_sources(
        sources=sources,
        include=config.lint.include,
        exclude=config.lint.exclude,
        respect_gitignore=config.respect_gitignore,
    )

    if add_file_level_general_noqa:
        sources_modified = noqa.add_file_level_general_ignore(included_sources)
        sys.stdout.write(
            f"File-level general noqa directive added to {sources_modified} file(s)\n",
        )

    # the `--workers` flag when provided, takes precedence over the environment variable
    # the environment variable when provided, takes precedence over the default
    workers = (
        workers
        if workers
        else int(os.getenv(WORKERS_ENVIRONMENT_VARIABLE, DEFAULT_WORKERS))
    )

    # we set the number of processes to the smallest of these values:
    # 1. the number of CPUs
    # 2. the number of workers
    with multiprocessing.Pool(
        processes=min(
            multiprocessing.cpu_count(),
            workers,
        ),
    ) as pool:
        try:
            results = [
                pool.apply_async(
                    linter.run,
                    kwds={
                        "source_file": source.resolve(),
                        "source_code": source.read_text(encoding="utf-8"),
                    },
                )
                for source in included_sources
            ]
            pool.close()
            pool.join()

            lint_results = [result.get() for result in results]

        except errors.ParseError as error:
            core.logger.error(error)
            sys.exit(2)

    for lint_result in lint_results:
        violations = linter.get_violation_stats(
            lint_result.violations,
        )

        linter.print_violations(
            violations=lint_result.violations,
            source_file=lint_result.source_file,
        )

        total_violations += violations.total
        auto_fixable_violations += violations.auto_fixable
        fix_enabled_violations += violations.fix_enabled

        if lint_result.fixed_sql:
            with pathlib.Path(lint_result.source_file).open(
                "w",
                encoding="utf-8",
            ) as sf:
                sf.write(lint_result.fixed_sql)

    if total_violations > 0:
        if config.lint.fix:
            sys.stdout.write(
                f"Found {total_violations} violation(s)"
                f" ({fix_enabled_violations} fixed,"
                f" {total_violations - fix_enabled_violations} remaining)\n",
            )

            if (total_violations - fix_enabled_violations) > 0:
                sys.exit(1)

        else:
            sys.stdout.write(
                f"Found {total_violations} violation(s)\n"
                f"{auto_fixable_violations} fix(es) available, {fix_enabled_violations} fix(es) enabled\n",  # noqa: E501
            )

            if auto_fixable_violations > 0:
                sys.stdout.write(
                    "Use with '--fix' to auto fix the violations\n",
                )

            sys.exit(1)


@cli.command(name="format")
@click.option(
    "--check",
    is_flag=True,
    default=False,
    help="Check if any files would have been modified.",
)
@click.option(
    "--diff",
    is_flag=True,
    default=False,
    help="""
    Report the difference between the current file and
    how the formatted file would look like.""",
)
@click.option(
    "--no-cache",
    is_flag=True,
    default=False,
    help="Whether to read the cache.",
)
@common_options
@click.argument("sources", nargs=-1, type=click.Path(exists=True, path_type=pathlib.Path))  # type: ignore [type-var]
def format_sources(  # noqa: C901, PLR0913
    sources: tuple[pathlib.Path, ...],
    *,
    check: bool,
    diff: bool,
    no_cache: bool,
    workers: int,
    verbose: bool,
) -> None:
    """Format SQL files.

    Parameters:
    ----------
    sources: tuple[pathlib.Path, ...]
        List of sources to format.
    check: bool
        Check if any files would have been modified.
    diff: bool
        Report the difference between the current file and
        how the formatted file would look like.
    no_cache: bool
        Whether to read the cache.
    workers: int
        Number of workers to use.
    verbose: bool
        Enable verbose logging.

    Returns:
    -------
    None

    """
    core.logger.setLevel(logging.INFO if verbose else logging.WARNING)

    console = Console()

    config: core.Config = core.parse_config()

    for key, value in [("check", check), ("diff", diff), ("no_cache", no_cache)]:
        if value:
            setattr(config.format, key, value)

    formatter: core.Formatter = core.Formatter(
        config=config,
        formatters=core.load_formatters,
    )

    # Use the current working directory if no sources are specified
    if not sources:
        sources = (pathlib.Path.cwd(),)

    included_sources = core.filter_sources(
        sources=sources,
        include=config.format.include,
        exclude=config.format.exclude,
        respect_gitignore=config.respect_gitignore,
    )

    cache = core.Cache(config=config)

    sources_to_reformat = included_sources

    if not config.format.no_cache:
        sources_to_reformat = cache.filter_sources(
            sources=included_sources,
        )

    changes_detected = False

    # the `--workers` flag when specified, takes precedence over the environment variable
    # the environment variable when provided, takes precedence over the default
    workers = (
        workers
        if workers
        else int(os.getenv(WORKERS_ENVIRONMENT_VARIABLE, DEFAULT_WORKERS))
    )

    # we set the number of processes to the smallest of these values:
    # 1. the number of CPUs
    # 2. the number of workers
    with multiprocessing.Pool(
        processes=min(
            multiprocessing.cpu_count(),
            workers,
        ),
    ) as pool:
        try:
            results = [
                pool.apply_async(
                    formatter.format,
                    kwds={
                        "source_file": source.resolve(),
                        "source_code": source.read_text(encoding="utf-8"),
                    },
                )
                for source in included_sources
            ]
            pool.close()
            pool.join()

            formatting_results = [result.get() for result in results]

        except (errors.ParseError, errors.MissingStatementTerminatorError) as error:
            core.logger.error(error)
            sys.exit(2)

    for formatting_result in formatting_results:
        if (
            formatting_result.formatted_source_code
            != formatting_result.original_source_code
            and not changes_detected
        ):
            changes_detected = True

        if config.format.diff:
            diff_unified = difflib.unified_diff(
                formatting_result.original_source_code.splitlines(keepends=True),
                formatting_result.formatted_source_code.splitlines(keepends=True),
                fromfile=str(formatting_result.source_file),
                tofile=str(formatting_result.source_file),
            )

            diff_output = "".join(diff_unified)

            if diff_output:
                console.print(Syntax(diff_output, "diff", theme="ansi_dark"))

        if not config.format.check and not config.format.diff:
            with pathlib.Path(formatting_result.source_file).open(
                "w",
                encoding="utf-8",
            ) as sf:
                sf.write(formatting_result.formatted_source_code)

    if not config.format.check and not config.format.diff:
        cache.write(sources=included_sources)
        sys.stdout.write(
            f"{len(sources_to_reformat)} file(s) reformatted, "
            f"{len(included_sources) - len(sources_to_reformat)} file(s) left unchanged\n",  # noqa: E501
        )

    if changes_detected and (config.format.check or config.format.diff):
        sys.exit(1)


if __name__ == "__main__":
    cli()



"""Primary key constraint creating index."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class PrimaryKeyConstraintCreatingIndex(linter.BaseChecker):
    """## **What it does**
    Checks primary key constraint creating index.

    ## **Why not?**
    Primary key constraint is supported by a unique index and a **NOT NULL** constraint
    in the background.
    Adding a primary key constraint to an already populated table will create a
    unique index in non-concurrent mode, scanning and validating that there are no
    violating records in the table. This causes the table to be locked, in which no other
    operations can be performed on the table for the duration of the validation. This will
    cause downtime if the table is concurrently being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    1. Create a unique index in concurrent mode:
        **CREATE UNIQUE INDEX CONCURRENTLY .. ON .. (..)**.
    2. In case the primary key column(s) are nullable, create a non-validating check
       constraint: **CHECK (column(s) IS NOT NULL) NOT VALID**.
    3. Validate the constraint.
    4. Set the primary key column(s) as **NOT NULL**.
    5. Drop the constraint.
    6. Add the primary key constraint using the index created in step 1:
        **ALTER TABLE .. ADD CONSTRAINT .. PRIMARY KEY USING INDEX ..**
    """

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> None:
        """Visit Constraint."""
        if (
            ancestors.find_nearest(ast.AlterTableCmd)
            and node.contype == enums.ConstrType.CONSTR_PRIMARY
            and not node.indexname
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Primary key constraint creating index",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Create a unique index in concurrent mode, have the primary key column(s) as NOT NULL, then add the primary key constraint using the index",  # noqa: E501
                ),
            )



"""Checker for numeric with precision."""

from pglast import ast, visitors

from pgrubic.core import linter


class NumericWithPrecision(linter.BaseChecker):
    """## **What it does**
    Checks for usage of numeric with precision.

    ## **Why not?**
    Because it rounds off the fractional part which can lead to rounding errors slipping
    in when performing calculations and storing partial results before aggregation.
    An example of what could go wrong is
    [how-to-prevent-postgresql-from-automatically-rounding-numeric-types](https://dba.stackexchange.com/questions/281953/how-to-prevent-postgresql-from-automatically-rounding-numeric-types){:target="_blank"}

    ## **When should you?**
    When you want to, really. If what you want is a numeric field that will throw an
    error when you insert too large a value into it, and you do not want to use an
    explicit check constraint and you always want to store the fractional part in a
    specific decimal places then numeric(p, s) is a perfectly good type.
    Just don't use it automatically without thinking about it.

    ## **Use instead:**
    numeric.
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if node.typeName.names[-1].sval == "numeric" and node.typeName.typmods:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Prefer entire numeric",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use numeric",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.typeName = ast.TypeName(
            names=(
                {
                    "@": "String",
                    "sval": "numeric",
                },
            ),
        )



"""Checker for nullable boolean field."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class NullableBooleanField(linter.BaseChecker):
    """## **What it does**
    Checks for usage of numeric with precision.

    ## **Why not?**
    3 possible values is not a boolean. By allowing nulls in a boolean field, you are
    turning an intended binary representation (true/false) into a ternary representation
    (true, false, null). Null is neither 'true' nor 'false'.
    Allowing nulls in a boolean field is an oversight that leads to unnecessarily
    ambiguous data.

    ## **When should you?**
    Never.

    ## **Use instead:**
    boolean with not null constraint.
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if node.typeName.names[-1].sval == "bool":
            is_not_null = bool(
                (
                    [
                        constraint
                        for constraint in node.constraints
                        if constraint.contype == enums.ConstrType.CONSTR_NOTNULL
                    ]
                    if node.constraints is not None
                    else []
                ),
            )

            if not is_not_null:
                self.violations.add(
                    linter.Violation(
                        rule_code=self.code,
                        rule_name=self.name,
                        rule_category=self.category,
                        line_number=self.line_number,
                        column_offset=self.column_offset,
                        line=self.line,
                        statement_location=self.statement_location,
                        description="Boolean field should not be nullable",
                        is_auto_fixable=self.is_fix_applicable,
                        is_fix_enabled=self.is_fix_enabled,
                        help="Add not null constraint",
                    ),
                )

                self._fix(node)

    def _fix(self, node: ast.ColumnDef) -> None:
        """Fix violation."""
        node.constraints = (
            *(node.constraints or []),
            ast.Constraint(
                contype=enums.ConstrType.CONSTR_NOTNULL,
            ),
        )



"""Checker for removal of required columns."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class RequiredColumnRemoval(linter.BaseChecker):
    """## **What it does**
    Checks for removal of required columns.

    ## **Why not?**
    If a column has been specified as required and you are removing it from a table,
    you are probably doing something wrong.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Leave the required column.

    ## **Configuration**
    `required-columns`: List of required columns along with their data types.
    """

    def visit_AlterTableCmd(
        self,
        ancestors: visitors.Ancestor,
        node: ast.AlterTableCmd,
    ) -> None:
        """Visit AlterTableCmd."""
        if node.subtype == enums.AlterTableType.AT_DropColumn:
            for column in self.config.lint.required_columns:
                if node.name == column.name:
                    self.violations.add(
                        linter.Violation(
                            rule_code=self.code,
                            rule_name=self.name,
                            rule_category=self.category,
                            line_number=self.line_number,
                            column_offset=self.column_offset,
                            line=self.line,
                            statement_location=self.statement_location,
                            description=f"Column `{node.name}` is marked as required"
                            " in config",
                            is_auto_fixable=self.is_auto_fixable,
                            is_fix_enabled=self.is_fix_enabled,
                            help="Leave the required column",
                        ),
                    )



"""Checker for null constraint."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class NullConstraint(linter.BaseChecker):
    """## **What it does**
    Checks for use of **NULL** constraint.

    ## **Why not?**
    From the documentation:

    The NOT NULL constraint has an inverse: the NULL constraint. This does not mean that
    the column must be null, which would surely be useless. Instead, this simply selects
    the default behavior that the column might be null. The NULL constraint is not present
    in the SQL standard and should not be used in portable applications.
    (It was only added to PostgreSQL to be compatible with some other database systems.)
    Since it is the default for any column, its presence is simply noise.

    ## **When should you?**
    Almost Never.

    ## **Use instead:**
    Leave out **NULL** constraints.
    """

    is_auto_fixable: bool = True

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> visitors.ActionMeta | None:
        """Visit Constraint."""
        if node.contype == enums.ConstrType.CONSTR_NULL:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="NULL constraints are redundant",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Remove the NULL constraint",
                ),
            )

            return self._fix()

        return None

    def _fix(self) -> visitors.ActionMeta:
        """Fix violation."""
        return visitors.Delete



"""Formatter for database."""

from pglast import ast, stream, printers

from pgrubic import Operators


@printers.node_printer(ast.CreatedbStmt, ast.DefElem, override=True)
def create_db_stmt_def_elem(node: ast.DefElem, output: stream.RawStream) -> None:
    """Printer for CreatedbStmt defelem."""
    option = node.defname
    if option == "connection_limit":
        output.write("CONNECTION LIMIT")
    else:
        output.write(node.defname.upper())
    if node.arg is not None:
        output.space()
        output.write(Operators.EQ)
        output.space()
        if isinstance(node.arg, tuple) or option in ("allow_connections", "is_template"):
            output.write(node.arg.sval)
        elif isinstance(node.arg, ast.String):
            if option.lower() in ("encoding", "locale", "strategy"):
                output.write_quoted_string(node.arg.sval.upper())
            else:
                output.write_quoted_string(node.arg.sval)
        else:
            output.print_node(node.arg)


@printers.node_printer(ast.DropdbStmt, override=True)
def drop_db_stmt(node: ast.DropdbStmt, output: stream.RawStream) -> None:
    """Printer for DropdbStmt."""
    output.write("DROP DATABASE")
    if node.missing_ok:
        output.space()
        output.write("IF EXISTS")
    output.space()
    output.print_name(node.dbname)
    if node.options:
        output.newline()
        output.write("WITH")
        output.space()
        with output.expression(need_parens=True):
            output.newline()
            output.space(4)
            output.print_list(node.options, "")
            output.newline()


@printers.node_printer(ast.DropdbStmt, ast.DefElem, override=True)
def drop_db_stmt_def_elem(node: ast.DefElem, output: stream.RawStream) -> None:
    """Printer for DropdbStmt defelem."""
    output.write(node.defname.upper())



"""Validating foreign key constraint on existing rows."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class ValidatingForeignKeyConstraintOnExistingRows(linter.BaseChecker):
    """## **What it does**
    Checks validating foreign key constraint on existing rows.

    ## **Why not?**
    Adding a foreign key constraint to an already populated table will have to scan and
    validate that there are no violating records in the table, causing the table to be
    locked in which no other operations can be performed on the table for the
    duration of the validation. This will cause downtime if the table is concurrently
    being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    1. Add the foreign key constraint without validation.
    2. Validate the constraint in a separate transaction.
    """

    is_auto_fixable = True

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> None:
        """Visit Constraint."""
        if (
            ancestors.find_nearest(ast.AlterTableCmd)
            and node.contype == enums.ConstrType.CONSTR_FOREIGN
            and not node.skip_validation
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Validating foreign key constraint on existing rows",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Add the foreign key constraint without validation and validate it in a separate transaction",  # noqa: E501
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.Constraint) -> None:
        """Fix violation."""
        node.skip_validation = True



"""Checker for single letter identifiers."""

from pgrubic.core import linter
from pgrubic.rules.naming import CheckIdentifier


class SingleLetterIdentifier(CheckIdentifier):
    """## **What it does**
    Checks for usage of single letter identifiers.

    ## **Why not?**
    Single letter identifier does not provide much information about what the identifier
    represents. Using a more descriptive name can improve clarity, readability, and
    maintainability of the database schema.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Descriptive name.
    """

    def _check_identifier(
        self,
        identifier: str,
        line_number: int,
        column_offset: int,
        line: str,
        statement_location: int,
    ) -> None:
        """Checks for identifiers with single letter."""
        if identifier and len(identifier.strip()) == 1:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=line_number,
                    column_offset=column_offset,
                    line=line,
                    statement_location=statement_location,
                    description=f"Single letter identifier `{identifier}`"
                    " is not descriptive enough",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use a more descriptive name",
                ),
            )



"""Methods to load rules and formatters."""

from __future__ import annotations

import os
import typing
import fnmatch
import inspect
import importlib

from pgrubic import (
    RULES_DIRECTORY,
    RULES_BASE_MODULE,
    FORMATTERS_DIRECTORY,
    FORMATTERS_BASE_MODULE,
)
from pgrubic.core import config, linter


def load_rules(config: config.Config) -> set[linter.BaseChecker]:
    """Load rules.

    Parameters:
    ----------
    config: config.Config
        Config.

    Returns:
    -------
    set[linter.BaseChecker]
        Set of rules.
    """
    rules: set[linter.BaseChecker] = set()

    for path in sorted(RULES_DIRECTORY.rglob("[!_]*.py"), key=lambda x: x.name):
        module_path = (
            str(RULES_BASE_MODULE / path.relative_to(RULES_DIRECTORY))
            .replace(".py", "")
            .replace(os.path.sep, ".")
        )

        module = importlib.import_module(module_path)

        for _, rule in inspect.getmembers(
            module,
            lambda x: inspect.isclass(x) and x.__module__ == module.__name__,  # noqa: B023
        ):
            if (
                issubclass(rule, linter.BaseChecker)
                and not rule.__name__.startswith(
                    "_",
                )
                and (
                    not config.lint.select
                    or any(
                        fnmatch.fnmatch(rule.code, pattern + "*")
                        for pattern in config.lint.select
                    )
                )
                and not any(
                    fnmatch.fnmatch(rule.code, pattern + "*")
                    for pattern in config.lint.ignore
                )
            ):
                rules.add(rule)

    return rules


def load_formatters() -> set[typing.Callable[[], None]]:
    """Load formatters.

    Returns:
    -------
    set[typing.Callable[[], None]]
        Set of formatters.
    """
    formatters: set[typing.Callable[[], None]] = set()

    for path in sorted(FORMATTERS_DIRECTORY.rglob("[!_]*.py"), key=lambda x: x.name):
        module_path = (
            str(FORMATTERS_BASE_MODULE / path.relative_to(FORMATTERS_DIRECTORY))
            .replace(".py", "")
            .replace(os.path.sep, ".")
        )

        module = importlib.import_module(module_path)

        for _, formatter in inspect.getmembers(
            module,
            lambda x: inspect.isfunction(x) and x.__module__ == module.__name__,  # noqa: B023
        ):
            formatters.add(formatter)

    return formatters



"""Test filters."""

import os
import pathlib
from unittest.mock import patch

import git

from pgrubic import core
from pgrubic.core import config


def test_filter_linting_sources(tmp_path: pathlib.Path) -> None:
    """Test filter linting sources."""
    directory = tmp_path / "sub"
    directory.mkdir()

    sql_fail: str = "SELECT a = NULL;"

    sources: tuple[pathlib.Path, ...] = (
        pathlib.Path("test.sql"),
        pathlib.Path("test_main.sql"),
        pathlib.Path("test.py"),
        pathlib.Path("test.txt"),
        pathlib.Path("tables.sql"),
        pathlib.Path("views.sql"),
    )

    for source in sources:
        file_fail = directory / source
        file_fail.write_text(sql_fail)

    expected_sources_filtered_length = 2

    config_content = f"""
    include = [
        "*.py",
    ]
    exclude = [
        "{str(directory).replace(os.path.sep, '/')}/test_main.sql",
    ]
    [lint]
    include = [
        "*.sql",
        "*.txt",
    ]
    exclude = [
        "{str(directory).replace(os.path.sep, '/')}/test.sql",
    ]
    """

    config_file = directory / config.CONFIG_FILE
    config_file.write_text(config_content)

    with patch.dict(
        "os.environ",
        {config.CONFIG_PATH_ENVIRONMENT_VARIABLE: str(directory)},
    ):
        parsed_config = config.parse_config()

        sources_filtered = core.filter_sources(
            sources=(directory,),
            include=parsed_config.lint.include,
            exclude=parsed_config.lint.exclude,
            respect_gitignore=parsed_config.respect_gitignore,
        )

        assert len(sources_filtered) == expected_sources_filtered_length


def test_filter_formatting_sources(tmp_path: pathlib.Path) -> None:
    """Test filter formatting sources."""
    directory = tmp_path / "sub"
    directory.mkdir()

    sql_fail: str = "SELECT a = NULL;"

    sources: tuple[pathlib.Path, ...] = (
        pathlib.Path("test.sql"),
        pathlib.Path("test_main.sql"),
        pathlib.Path("test.py"),
        pathlib.Path("test.txt"),
        pathlib.Path("tables.sql"),
        pathlib.Path("views.sql"),
    )

    for source in sources:
        file_fail = directory / source
        file_fail.write_text(sql_fail)

    expected_sources_filtered_length = 2

    config_content = f"""
    include = [
        "*.py",
    ]
    exclude = [
        "{str(directory).replace(os.path.sep, '/')}/test_main.sql",
    ]
    [format]
    include = [
        "*.sql",
        "*.txt",
    ]
    exclude = [
        "{str(directory).replace(os.path.sep, '/')}/test.sql",
    ]
    """

    config_file = directory / config.CONFIG_FILE
    config_file.write_text(config_content)

    with patch.dict(
        "os.environ",
        {config.CONFIG_PATH_ENVIRONMENT_VARIABLE: str(directory)},
    ):
        parsed_config = config.parse_config()

        sources_filtered = core.filter_sources(
            sources=(directory,),
            include=parsed_config.format.include,
            exclude=parsed_config.format.exclude,
            respect_gitignore=parsed_config.respect_gitignore,
        )

        assert len(sources_filtered) == expected_sources_filtered_length


def test_respect_gitignore_filter_sources(tmp_path: pathlib.Path) -> None:
    """Test respect gitignore filter sources."""
    # Disable global git config
    os.environ["GIT_CONFIG_GLOBAL"] = "/dev/null"

    git.Repo.init(tmp_path)

    # Create a .gitignore file in the repository
    gitignore_file = tmp_path / ".gitignore"
    gitignore_file.write_text("ignored_file.sql\n")

    directory = tmp_path / "sub"
    directory.mkdir()

    sql_fail: str = "SELECT a = NULL;"

    sources: tuple[pathlib.Path, ...] = (
        pathlib.Path("test.sql"),
        pathlib.Path("test_main.sql"),
        pathlib.Path("test.py"),
        pathlib.Path("test.txt"),
        pathlib.Path("ignored_file.sql"),
        pathlib.Path("tables.sql"),
        pathlib.Path("views.sql"),
    )

    for source in sources:
        file_fail = directory / source
        file_fail.write_text(sql_fail)

    expected_sources_filtered_length = 2

    config_content = f"""
    include = [
        "*.sql",
    ]
    exclude = [
        "{str(directory).replace(os.path.sep, '/')}/test_main.sql",
        "{str(directory).replace(os.path.sep, '/')}/test.sql",
    ]
    """

    config_file = directory / config.CONFIG_FILE
    config_file.write_text(config_content)

    with patch.dict(
        "os.environ",
        {config.CONFIG_PATH_ENVIRONMENT_VARIABLE: str(directory)},
    ):
        parsed_config = config.parse_config()

        sources_filtered = core.filter_sources(
            sources=(directory,),
            include=parsed_config.include,
            exclude=parsed_config.exclude,
            respect_gitignore=parsed_config.respect_gitignore,
        )

        assert len(sources_filtered) == expected_sources_filtered_length


def test_respect_gitignore_false_filter_sources(tmp_path: pathlib.Path) -> None:
    """Test respect gitignore false filter sources."""
    # Disable global git config
    os.environ["GIT_CONFIG_GLOBAL"] = "/dev/null"

    git.Repo.init(tmp_path)

    # Create a .gitignore file in the repository
    gitignore_file = tmp_path / ".gitignore"
    gitignore_file.write_text("ignored_file.sql\n")

    directory = tmp_path / "sub"
    directory.mkdir()

    sql_fail: str = "SELECT a = NULL;"

    sources: tuple[pathlib.Path, ...] = (
        pathlib.Path("test.sql"),
        pathlib.Path("test_main.sql"),
        pathlib.Path("test.py"),
        pathlib.Path("test.txt"),
        pathlib.Path("ignored_file.sql"),
        pathlib.Path("tables.sql"),
        pathlib.Path("views.sql"),
    )

    for source in sources:
        file_fail = directory / source
        file_fail.write_text(sql_fail)

    expected_sources_filtered_length = 3

    config_content = f"""
    respect-gitignore = false
    include = [
        "*.sql",
    ]
    exclude = [
        "{str(directory).replace(os.path.sep, '/')}/test_main.sql",
        "{str(directory).replace(os.path.sep, '/')}/test.sql",
    ]
    """

    config_file = directory / config.CONFIG_FILE
    config_file.write_text(config_content)

    with patch.dict(
        "os.environ",
        {config.CONFIG_PATH_ENVIRONMENT_VARIABLE: str(directory)},
    ):
        parsed_config = config.parse_config()

        sources_filtered = core.filter_sources(
            sources=(directory,),
            include=parsed_config.include,
            exclude=parsed_config.exclude,
            respect_gitignore=parsed_config.respect_gitignore,
        )

        assert len(sources_filtered) == expected_sources_filtered_length



"""Checker for non snake case identifiers."""

from pglast import stream

from pgrubic.core import linter
from pgrubic.rules.naming import CheckIdentifier


class NonSnakeCaseIdentifier(CheckIdentifier):
    """## **What it does**
    Check if identifier is not in snake case.

    ## **Why not?**
    PostgreSQL folds all names - of tables, columns, functions and everything else - to
    lower case unless they're "double quoted".
    So `create table Foo()` will create a table called foo, while `create table "Bar"()`
    will create a table called Bar.

    These select commands will work: `select * from Foo`, `select * from foo`,
    `select * from "Bar"`.

    These will fail with "no such table": `select * from "Foo"`, `select * from Bar`,
    `select * from bar`.

    This means that if you use uppercase characters in your table or column names you
    have to either always double quote them or never double quote them.
    That's annoying enough by hand, but when you start using other tools to access the
    database, some of which always quote all names and some don't, it gets very confusing.

    ## **When should you?**
    Never ... almost.
    If it is important that "pretty" names are displaying in report output then you might
    want to use them. But you can also use column aliases to use lower case names in a
    table and still get pretty names in the output of a query:
    > select character_name as "Character Name" from foo.

    ## **Use instead:**
    Stick to using **a-z, 0-9 and underscore** for names and you never have to worry about
    quoting them.
    """

    def _check_identifier(
        self,
        identifier: str,
        line_number: int,
        column_offset: int,
        line: str,
        statement_location: int,
    ) -> None:
        """Check if identifier is not in snake case."""
        if identifier and not stream.is_simple_name(identifier):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=line_number,
                    column_offset=column_offset,
                    line=line,
                    statement_location=statement_location,
                    description=f"Identifier `{identifier}` should be in snake case",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use snake case for identifiers",
                ),
            )



"""Checker for not null constraint on new column with volatile default."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class NotNullConstraintOnNewColumnWithVolatileDefault(linter.BaseChecker):
    """## **What it does**
    Checks **NOT NULL** constraint on a new column with volatile default.

    ## **Why not?**
    Adding a new column with **NOT NULL** constraint and a volatile default to an already
    populated table will have to backfill the newly added column with the default, causing
    the table to be locked, in which no other operations can be performed on the table for
    the duration of the backfill. This will cause downtime if the table is concurrently
    being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    1. Create the new column, nullable.
    2. Set the default value of the new column to the volatile default.
    3. Backfill the new column for all existing rows.
    4. Create a check constraint: **CHECK (column IS NOT NULL) NOT VALID**.
    5. Validate the constraint.
    6. Set the column as NOT NULL.
    7. Drop the constraint.
    """

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if ancestors.find_nearest(ast.AlterTableCmd) and node.constraints:
            is_not_null = False
            has_static_default = False

            for constraint in node.constraints:
                if constraint.contype == enums.ConstrType.CONSTR_NOTNULL:
                    is_not_null = True

                if constraint.contype == enums.ConstrType.CONSTR_DEFAULT and isinstance(
                    constraint.raw_expr,
                    ast.A_Const,
                ):
                    has_static_default = True

            if is_not_null and not has_static_default:
                self.violations.add(
                    linter.Violation(
                        rule_code=self.code,
                        rule_name=self.name,
                        rule_category=self.category,
                        line_number=self.line_number,
                        column_offset=self.column_offset,
                        line=self.line,
                        statement_location=self.statement_location,
                        description="Not null constraint on new column with volatile default",  # noqa: E501
                        is_auto_fixable=self.is_auto_fixable,
                        is_fix_enabled=self.is_fix_enabled,
                        help="Split the operation into multiple steps",
                    ),
                )



"""Checker for extension whitelist."""

from pglast import ast, visitors

from pgrubic.core import linter


class ExtensionWhitelist(linter.BaseChecker):
    """## **What it does**
    Checks that an extension to be created is allowed.

    ## **Why not?**
    By default, any extension can be loaded into the database.
    This is quite dangerous as any bug causing a crash would mean a PostgreSQL would
    restart. So you not only want to empower **CREATE EXTENSION** to database owners,
    you also want to be able to review and explicitly allow extensions.

    ## **When should you?**
    Almost never. If an extension is not allowed, you are probably doing
    something wrong.

    ## **Use instead:**
    Extensions that are allowed.

    ## **Configuration**
    `allowed-extensions`: List of allowed extensions.
    """

    def visit_CreateExtensionStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateExtensionStmt,
    ) -> None:
        """Visit CreateExtensionStmt."""
        if (
            node.extname not in self.config.lint.allowed_extensions
            and "*" not in self.config.lint.allowed_extensions
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Extension '{node.extname}' is not allowed",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="If you want to allow this extension, add the extension to the"
                    " allowed extensions list",
                ),
            )



"""Checker for DELETE without a WHERE clause."""

from pglast import ast, visitors

from pgrubic.core import linter


class DeleteWithoutWhereClause(linter.BaseChecker):
    """## **What it does**
    Checks for **DELETE** without a **WHERE** clause.

    ## **Why not?**
    Executing a **DELETE** line without a **WHERE** clause will remove all the rows
    in the table which is most likely an accidental mistake and not what you want.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Add necessary **WHERE** clause.
    """

    def visit_DeleteStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.DeleteStmt,
    ) -> None:
        """Visit DeleteStmt."""
        if not node.whereClause:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Found DELETE without a WHERE clause",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Add WHERE clause",
                ),
            )



"""Tests for config."""

import pathlib
from unittest.mock import patch

import pytest

from pgrubic.core import config, errors


def test_config_from_environment_variable(tmp_path: pathlib.Path) -> None:
    """Test config from environment variable."""
    config_content = """
    [lint]
    fix = true
    """
    directory = tmp_path / "sub"
    directory.mkdir()

    config_file = directory / config.CONFIG_FILE
    config_file.write_text(config_content)

    with patch.dict(
        "os.environ",
        {config.CONFIG_PATH_ENVIRONMENT_VARIABLE: str(directory)},
    ):
        parsed_config = config.parse_config()
        assert parsed_config.lint.fix is True


def test_config_from_current_working_directory(tmp_path: pathlib.Path) -> None:
    """Test config from current working directory."""
    config_content = """
    [format]
    diff = true
    """
    directory = tmp_path / "sub"
    directory.mkdir()

    config_file = directory / config.CONFIG_FILE
    config_file.write_text(config_content)

    with patch("os.getcwd", return_value=str(directory)):
        assert pathlib.Path.cwd() == directory

        parsed_config = config.parse_config()
        assert parsed_config.format.diff is True


def test_missing_config_error(tmp_path: pathlib.Path) -> None:
    """Test missing config error."""
    config_content = """
    [lint]
    required-columns = [
        { name = "foo", type = "text" }
    ]
    """
    directory = tmp_path / "sub"
    directory.mkdir()

    config_file = directory / config.CONFIG_FILE
    config_file.write_text(config_content)

    with patch("os.getcwd", return_value=str(directory)):
        assert pathlib.Path.cwd() == directory

        with pytest.raises(errors.MissingConfigError) as excinfo:
            config.parse_config()

        assert excinfo.value.args[0] == "Missing config key: data-type"



"""Checker for adding of auto increment column."""

from pglast import ast, visitors

from pgrubic.core import linter


class AddingAutoIncrementColumn(linter.BaseChecker):
    """## **What it does**
    Checks adding of auto increment column.

    ## **Why not?**
    Adding an auto increment column to an already populated table will
    have to backfill the newly added column, causing the table to be locked
    in which no other operations can be performed on the table for the duration
    of the backfill. This will cause downtime if the table is concurrently
    being accessed by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently accessed.

    ## **Use instead:**
    1. Create a new column typed bigint, nullable.
    2. Create a sequence.
    3. Set the next value of the sequence to the total number of rows in the table
       with enough offset.
    4. Set the default value of the new column to the next value of the sequence.
    5. Backfill the new column for all existing rows.
    6. If the new column is to be set as not null,
       add a check constraint: **CHECK (column IS NOT NULL) NOT VALID**
    7. Validate the constraint.
    8. Set the column as NOT NULL.
    9. Drop the constraint.
    """

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if ancestors.find_nearest(ast.AlterTableCmd) and (
            node.typeName.names[-1].sval in ["smallserial", "serial", "bigserial"]
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Adding auto increment column is not safe",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Split the operation into multiple steps",
                ),
            )



"""Checker for column rename."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class ColumnRename(linter.BaseChecker):
    """## **What it does**
    Checks renaming of column.

    ## **Why not?**
    Renaming a column can easily break applications that rely on the column.

    If any part of the application code, database procedures, views, or reports use
    the column, renaming it will cause errors and potentially disrupt business operations.

    ## **When should you?**
    If the column is no longer being referenced by clients, probably after migrating
    clients to a new column.

    ## **Use instead:**
    1. Create a new column with the new name, nullable.
    2. Start writing data to the new column.
    3. Copy all data from the old column to the new column.
    4. Migrate clients to the new column.
    """

    def visit_RenameStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.RenameStmt,
    ) -> None:
        """Visit RenameStmt."""
        if (
            node.renameType == enums.ObjectType.OBJECT_COLUMN
            and node.newname != node.subname
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Column rename is not safe",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Create a new column with the new name",
                ),
            )



"""Formatter for UPDATE statements."""

from pglast import ast, stream, printers


@printers.node_printer(ast.UpdateStmt, override=True)
def update_stmt(node: ast.UpdateStmt, output: stream.RawStream) -> None:
    """Printer for UpdateStmt."""
    with output.push_indent():
        if node.withClause:
            output.write("WITH")
            output.space()
            output.print_node(node.withClause)
            output.indent()

        output.write("UPDATE")
        output.space()
        output.print_node(node.relation)
        output.newline()
        output.space(3)
        output.write("SET")
        output.space()
        output.print_list(node.targetList, standalone_items=False)

        if node.fromClause:
            output.newline()
            output.space(2)
            output.write("FROM")
            output.space()
            output.print_list(node.fromClause)

        if node.whereClause:
            output.newline()
            output.space()
            output.write("WHERE")
            output.space()
            output.print_node(node.whereClause)

            if node.returningList:
                output.newline()
                output.write("RETURNING")
                output.space()
                output.print_list(node.returningList)

        if node.withClause:
            output.dedent()



"""Checker for missing required columns."""

from pglast import ast, enums, visitors

from pgrubic.core import config, linter
from pgrubic.rules.general import get_columns_from_table_creation


class MissingRequiredColumn(linter.BaseChecker):
    """## **What it does**
    Checks for missing required column.

    ## **Why not?**
    If a column has been specified as required and you have not defined it,
    you are probably doing something wrong.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Define the required column.

    ## **Configuration**
    `required-columns`: List of required columns along with their data types.
    """

    is_auto_fixable: bool = True

    def visit_CreateStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateStmt,
    ) -> None:
        """Visit CreateStmt."""
        if not node.partbound and node.tableElts:
            given_columns, _ = get_columns_from_table_creation(node)

            for required_column in self.config.lint.required_columns:
                if not any(
                    required_column.name == column.name for column in given_columns
                ):
                    self.violations.add(
                        linter.Violation(
                            rule_code=self.code,
                            rule_name=self.name,
                            rule_category=self.category,
                            line_number=self.line_number,
                            column_offset=self.column_offset,
                            line=self.line,
                            statement_location=self.statement_location,
                            description=f"Column `{required_column.name}` of type"
                            f" `{required_column.data_type}` is marked as required in config",  # noqa: E501
                            is_auto_fixable=self.is_auto_fixable,
                            is_fix_enabled=self.is_fix_enabled,
                            help="Define the required column",
                        ),
                    )

                    self._fix(node, required_column)

    def _fix(self, node: ast.CreateStmt, column: config.Column) -> None:
        """Fix violation."""
        node.tableElts = (
            *node.tableElts,
            ast.ColumnDef(
                colname=column.name,
                typeName=ast.TypeName(
                    names=(
                        {
                            "@": "String",
                            "sval": column.data_type,
                        },
                    ),
                ),
                constraints=(
                    *(node.constraints or []),
                    ast.Constraint(
                        contype=enums.ConstrType.CONSTR_NOTNULL,
                    ),
                ),
            ),
        )



"""Checker for identifiers with special characters."""

from pgrubic.core import linter
from pgrubic.rules.naming import CheckIdentifier


class SpecialCharacterInIdentifier(CheckIdentifier):
    """## **What it does**
    Checks for identifiers with special characters.

    ## **Why not?**
    SQL identifiers must begin with a letter (a-z) or an underscore (_). Subsequent
    characters in a name can be letters, digits (0-9), or underscores.

    PostgreSQL won't allow special characters in identifiers without double quotes.
    This means that if you use special characters in identifiers, you have to
    always double quote them. That is annoying enough by hand and error-prone.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Remove special characters from the identifier.
    """

    def _check_identifier(
        self,
        identifier: str,
        line_number: int,
        column_offset: int,
        line: str,
        statement_location: int,
    ) -> None:
        """Checks for identifiers with special characters."""
        if identifier and not identifier.replace("_", "").isalnum():
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=line_number,
                    column_offset=column_offset,
                    line=line,
                    statement_location=statement_location,
                    description=f"Special characters in identifier `{identifier}`",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Remove special characters from the identifier",
                ),
            )



"""Checker for drop cascade."""

from pglast import ast, enums, visitors

from pgrubic import get_fully_qualified_name
from pgrubic.core import linter


class DropCascade(linter.BaseChecker):
    """## **What it does**
    Checks for usage of cascade in drop statements.

    ## **Why not?**
    Database schema should follow the principle of least surprise which
    says that every component in a system should behave in a way that most users
    expect it to behave, and therefore not surprise or astonish them.

    Cascading drops should not cause unexpected loss of data. It is certainly
    dangerous if dropping a single table can wipe out half your database.

    Are you certain you want cascade drop thus dropping every dependent objects?

    ## **When should you?**
    Almost never.

    ## **Use instead:**
    Remove the **cascade** or use **restrict**
    """

    is_auto_fixable: bool = True

    def _register_violation(
        self,
        object_name: str,
        line_number: int,
        column_offset: int,
        line: str,
        statement_location: int,
    ) -> None:
        """Register violation."""
        self.violations.add(
            linter.Violation(
                rule_code=self.code,
                rule_name=self.name,
                rule_category=self.category,
                line_number=line_number,
                column_offset=column_offset,
                line=line,
                statement_location=statement_location,
                description=f"Drop cascade on `{object_name}` detected",
                is_auto_fixable=self.is_auto_fixable,
                is_fix_enabled=self.is_fix_enabled,
                help="Remove the cascade or use restrict",
            ),
        )

    def visit_DropStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.DropStmt,
    ) -> None:
        """Visit DropStmt."""
        for obj in node.objects:
            object_names = getattr(obj, "names", getattr(obj, "objname", obj))

            if node.behavior == enums.DropBehavior.DROP_CASCADE:
                self._register_violation(
                    object_name=get_fully_qualified_name(object_names),
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                )

                self._fix_drop(node)

    def _fix_drop(self, node: ast.DropStmt) -> None:
        """Fix violation."""
        node.behavior = enums.DropBehavior.DROP_RESTRICT

    def visit_AlterTableCmd(
        self,
        ancestors: visitors.Ancestor,
        node: ast.AlterTableCmd,
    ) -> None:
        """Visit AlterTableCmd."""
        if (
            node.subtype
            in (
                enums.AlterTableType.AT_DropConstraint,
                enums.AlterTableType.AT_DropColumn,
            )
            and node.behavior == enums.DropBehavior.DROP_CASCADE
        ):
            self._register_violation(
                object_name=node.name,
                line_number=self.line_number,
                column_offset=self.column_offset,
                line=self.line,
                statement_location=self.statement_location,
            )

            self._fix_alter(node)

    def _fix_alter(self, node: ast.AlterTableCmd) -> None:
        """Fix violation."""
        node.behavior = enums.DropBehavior.DROP_RESTRICT



"""Configuration."""

import os
import typing
import pathlib
import dataclasses

import toml
from deepmerge import always_merger

from pgrubic import PACKAGE_NAME
from pgrubic.core import errors
from pgrubic.core.logger import logger

CONFIG_FILE: typing.Final[str] = f"{PACKAGE_NAME}.toml"

DEFAULT_CONFIG: typing.Final[pathlib.Path] = (
    pathlib.Path(__file__).resolve().parent.parent / CONFIG_FILE
)

CONFIG_PATH_ENVIRONMENT_VARIABLE: typing.Final[str] = (
    f"{PACKAGE_NAME.upper()}_CONFIG_PATH"
)


@dataclasses.dataclass(kw_only=True, frozen=True)
class DisallowedSchema:
    """Representation of disallowed schema."""

    name: str
    reason: str
    use_instead: str


@dataclasses.dataclass(kw_only=True, frozen=True)
class DisallowedDataType:
    """Representation of disallowed data type."""

    name: str
    reason: str
    use_instead: str


@dataclasses.dataclass(kw_only=True, frozen=True)
class Column:
    """Representation of column."""

    name: str
    data_type: str


@dataclasses.dataclass(kw_only=True)
class Lint:
    # fmt: off
    """
### **postgres-target-version**
The target version of Postgres to lint against. This is used to either enable or
disable certain linting rules. For example, `DETACH PARTITION CONCURRENTLY`
was introduced from Postgres 14.

**Type**: `int`

**Default**: `14`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
postgres-target-version = 12
```
</details>

### **select**
List of rule aliases or prefixes to enable. It can be the exact code of a rule or
an entire category of rules, for example, `TP017`, `TP`. All rules are enabled by default.
Can be used in combination with `ignore` to streamline rules selection.

**Type**: `list[str]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
select = ["TP"]
```
</details>

### **ignore**
List of rule aliases or prefixes to disable. It can be the exact code of a rule or
an entire category of rules, for example, `TP017`, `TP`.
Can be used in combination with `select` to streamline rules selection.
Please note that **ignore** takes precedence over **select**.

**Type**: `list[str]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
ignore = ["TP017"]
```
</details>

### **include**
List of file patterns to include in the linting process.

**Type**: `list[str]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
include = ["V*.sql"]
```
</details>

### **exclude**
List of file patterns to exclude from the linting process.

**Type**: `list[str]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
exclude = ["test*.sql"]
```
</details>

### **ignore-noqa**
Whether to ignore `NOQA` directives in sources.
Overridden by the `--ignore-noqa` command-line flag.

**Type**: `bool`

**Default**: `False`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
ignore-noqa = true
```
</details>

### **allowed-extensions**
List of allowed postgres extensions.

**Type**: `list[str]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
allowed-extensions = ["pg_stat_statements"]
```
</details>

### **allowed-languages**
List of allowed languages.

**Type**: `list[str]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
allowed-languages = ["plpgsql"]
```
</details>

### **required-columns**
List of required columns along with their data types for every table.

**Type**: `list[Column]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
required-columns = [
    { name = "created_at", data_type = "timestamptz" },
    { name = "updated_at", data_type = "timestamptz" },
]
```
</details>

### **disallowed-schemas**
List of disallowed schemas, with reasons for their disallowance and what to use
instead.

**Type**: `list[DisallowedSchema]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
disallowed-schemas = [
    { name = "public", reason = "public schema", use_instead = "app" },
]
```
</details>

### **disallowed-data-types**
List of disallowed data types, with reasons for their disallowance
and what to use instead.

**Type**: `list[DisallowedDataType]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
disallowed-data-types = [
    { name = "varchar", reason = "text is better", use_instead = "text" },
]
```
</details>

### **fix**
Whether to automatically fix fixable violations.
Overridden by the `--fix` command-line flag.

**Type**: `bool`

**Default**: `False`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
fix = true
```
</details>

### **fixable**
List of rule aliases or prefixes to consider fixable. It can be the exact code of a rule
or an entire category of rules, for example, `TP017`, `TP`. All rules are considered
fixable by default. Please note that **unfixable** takes precedence over **fixable**.

**Type**: `list[str]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
fixable = ["TP"]
```
</details>

### **unfixable**
List of rule aliases or prefixes to consider unfixable. It can be the exact code of a rule
or an entire category of rules, for example, `TP017`, `TP`.

**Type**: `list[str]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
unfixable = ["TP017"]
```
</details>

### **timestamp-column-suffix**
Suffix to add to timestamp columns.

**Type**: `str`

**Default**: `"_at"`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
timestamp-column-suffix = "_at"
```
</details>

### **date-column-suffix**
Suffix to add to date columns.

**Type**: `str`

**Default**: `"_on"`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
date-column-suffix = "_date"
```
</details>

### **regex-partition**
Regular expression to match partition names.

**Type**: `str`

**Default**: `r"^[a-z0-9_]+$"`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
regex-partition = r"^[a-z0-9_]+$"
```
</details>

### **regex-index**
Regular expression to match index names.

**Type**: `str`

**Default**: `r"^[a-z0-9_]+$"`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
regex-index = r"^[a-z0-9_]+$"
```
</details>

### **regex-constraint-primary-key**
Regular expression to match primary key constraint names.

**Type**: `str`

**Default**: `r"^[a-z0-9_]+$"`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
regex-constraint-primary-key = r"^[a-z0-9_]+$"
```
</details>

### **regex-constraint-unique-key**
Regular expression to match unique key constraint names.

**Type**: `str`

**Default**: `r"^[a-z0-9_]+$"`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
regex-constraint-unique-key = r"^[a-z0-9_]+$"
```
</details>

### **regex-constraint-foreign-key**
Regular expression to match foreign key constraint names.

**Type**: `str`

**Default**: `r"^[a-z0-9_]+$"`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
regex-constraint-foreign-key = r"^[a-z0-9_]+$"
```
</details>

### **regex-constraint-check**
Regular expression to match check constraint names.

**Type**: `str`

**Default**: `r"^[a-z0-9_]+$"`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
regex-constraint-check = r"^[a-z0-9_]+$"
```
</details>

### **regex-constraint-exclusion**
Regular expression to match exclusion constraint names.

**Type**: `str`

**Default**: `r"^[a-z0-9_]+$"`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
regex-constraint-exclusion = r"^[a-z0-9_]+$"
```
</details>

### **regex-sequence**
Regular expression to match sequence names.

**Type**: `str`

**Default**: `r"^[a-z0-9_]+$"`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[lint]
regex-sequence = r"^[a-z0-9_]+$"
```
</details>
    """  # noqa: D212, D207 # fmt: on

    postgres_target_version: int
    select: list[str]
    ignore: list[str]
    include: list[str]
    exclude: list[str]
    ignore_noqa: bool
    allowed_extensions: list[str]
    allowed_languages: list[str]
    required_columns: list[Column]
    disallowed_schemas: list[DisallowedSchema]
    disallowed_data_types: list[DisallowedDataType]

    fix: bool
    fixable: list[str]
    unfixable: list[str]

    timestamp_column_suffix: str
    date_column_suffix: str
    regex_partition: str
    regex_index: str
    regex_constraint_primary_key: str
    regex_constraint_unique_key: str
    regex_constraint_foreign_key: str
    regex_constraint_check: str
    regex_constraint_exclusion: str
    regex_sequence: str


@dataclasses.dataclass(kw_only=True)
class Format:
    # fmt: off
    """
### **include**
A list of file patterns to include in the formatting process.

**Type**: `list[str]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[format]
include = ["V*.sql"]
```
</details>

### **exclude**
A list of file patterns to exclude from the formatting process.

**Type**: `list[str]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[format]
exclude = ["test*.sql"]
```
</details>

### **comma-at-beginning**
If `true`, add comma as a prefix as opposed to a suffix when formatting a list of
items, such as list of columns in which each column is on a separate line.

For example, when `true`:
```sql
select column1
     , column2
     , column3
     , .......
```

when `false`:
```sql
select column1,
       column2,
       column3,
       .......
```

**Type**: `bool`

**Default**: `true`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[format]
comma-at-beginning = false
```
</details>

### **new-line-before-semicolon**
If `true`, add a new line before each semicolon.

For example, when `true`:
```sql
select column1
        , column2
        , column3
    from table
;
```

when `false`:
```sql
select column1,
        column2,
        column3
    from table;
```

**Type**: `bool`

**Default**: `false`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[format]
new-line-before-semicolon = true
```
</details>

### **lines-between-statements**
Number of lines between SQL statements.

**Type**: `int`

**Default**: `1`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[format]
lines-between-statements = 2
```
</details>

### **remove-pg-catalog-from-functions**
If `true`, remove the `pg_catalog.` prefix from functions. Postgres standard functions
are located in the `pg_catalog` schema and thus prefixed with `pg_catalog.`
by default.

**Type**: `bool`

**Default**: `true`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[format]
remove-pg-catalog-from-functions = false
```
</details>

### **diff**
When `true`, report the difference between the current file and how it will look when
formatted, without making any changes to the file. If there is a difference, it exits
with a non-zero exit code.

Overridden by the `--diff` command-line flag.

**Type**: `bool`

**Default**: `false`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[format]
diff = true
```
</details>

### **check**
When `true`, it exits with a non-zero exit code if the any files would have been
modified by the formatter.

Overridden by the `--check` command-line flag.

**Type**: `bool`

**Default**: `false`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[format]
check = true
```
</details>

### **no-cache**
Whether to read the cache. Caching helps speed up the formatting process. When a file
has not been modified after the last formatting, it is simply skipped.
To force reformatting of a file even if it has not been modified since the last
formatting, set to `true`.

Overridden by the `--no-cache` command-line flag.

**Type**: `bool`

**Default**: `false`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
[format]
no-cache = true
```
</details>
    """  # noqa: D212, D207 # fmt: on

    include: list[str]
    exclude: list[str]
    comma_at_beginning: bool
    new_line_before_semicolon: bool
    lines_between_statements: int
    remove_pg_catalog_from_functions: bool
    diff: bool
    check: bool
    no_cache: bool


@dataclasses.dataclass(kw_only=True)
class Config:
    # fmt: off
    """
### **cache-dir**
Path to the cache directory.

If default and the environment variable `PGRUBIC_CACHE` is set, the environment
variable takes precedence or otherwise the non-default set value is always used.

**Type**: `str`

**Default**: `".pgrubic_cache"`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
cache-dir = "~/.cache/pgrubic"
```
</details>

### **include**
A list of file patterns to include in the linting and formatting process.

**Type**: `list[str]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
include = ["V*.sql"]
```
</details>

### **exclude**
A list of file patterns to exclude from the linting and formatting process.

**Type**: `list[str]`

**Default**: `[]`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
exclude = ["test*.sql"]
```
</details>

### **respect-gitignore**
Whether to automatically exclude files that are ignored by `.ignore`, `.gitignore`,
`.git/info/exclude`, and global gitignore files. Enabled by default.

**Type**: `bool`

**Default**: `True`

**Example**:
<details open>
<summary><strong>pgrubic.toml</strong></summary>

```toml
respect-gitignore = false
```
</details>
    """  # noqa: D212, D207 # fmt: on

    cache_dir: pathlib.Path

    include: list[str]
    exclude: list[str]
    respect_gitignore: bool

    lint: Lint
    format: Format


def _load_default_config() -> dict[str, typing.Any]:
    """Load default config.

    Returns:
    -------
    dict[str, typing.Any]
        The default config.
    """
    return dict(toml.load(DEFAULT_CONFIG))


def _load_user_config() -> dict[str, typing.Any]:
    """Load config from absolute path config file.

    Returns:
    -------
    dict[str, typing.Any]
        The config from the absolute path config file.
    """
    config_file_absolute_path = _get_config_file_absolute_path()

    if config_file_absolute_path:
        return dict(toml.load(config_file_absolute_path))

    return {}  # pragma: no cover


def _merge_config() -> dict[str, typing.Any]:
    """Merge default and user config.

    Returns:
    -------
    dict[str, typing.Any]
        The merged config.
    """
    return dict(always_merger.merge(_load_default_config(), _load_user_config()))


def _get_config_file_absolute_path(
    config_file: str = CONFIG_FILE,
) -> pathlib.Path | None:
    """Get the absolute path of the config file.
    If CONFIG_PATH_ENVIRONMENT_VARIABLE environment variable is set, we try to use that
    else, we use the first config file that we find upwards from the current working
    directory.

    Returns:
    -------
    pathlib.Path | None
        The absolute path of the config file if found, else None.
    """
    env_config_path = os.getenv(CONFIG_PATH_ENVIRONMENT_VARIABLE)

    if env_config_path:
        config_file_absolute_path = pathlib.Path(env_config_path).resolve() / config_file
        if pathlib.Path.exists(config_file_absolute_path):
            logger.info(
                """Using settings from "%s\"""",
                config_file_absolute_path,
            )
            return config_file_absolute_path

    current_directory = pathlib.Path.cwd()

    # Traverse upwards through the directory tree
    while current_directory != current_directory.parent:
        # Check if the configuration file exists
        config_file_absolute_path = current_directory / config_file

        if pathlib.Path.exists(config_file_absolute_path):
            logger.info(
                """Using settings from "%s\"""",
                config_file_absolute_path,
            )
            return config_file_absolute_path

        # Move up one directory
        current_directory = current_directory.parent  # pragma: no cover

    logger.info(
        """Using default settings""",
    )

    return None  # pragma: no cover


def parse_config() -> Config:
    """Parse config.

    Returns:
    -------
    Config
        The parsed config.

    Raises:
    ------
    MissingConfigError
        Raised when a config is missing.
    """
    merged_config = _merge_config()
    config_lint = merged_config["lint"]
    config_format = merged_config["format"]

    try:
        return Config(
            cache_dir=pathlib.Path(merged_config["cache-dir"]),
            include=merged_config["include"],
            exclude=merged_config["exclude"],
            respect_gitignore=merged_config["respect-gitignore"],
            lint=Lint(
                postgres_target_version=config_lint["target-postgres-version"],
                select=config_lint["select"],
                ignore=config_lint["ignore"],
                include=config_lint["include"] + merged_config["include"],
                exclude=config_lint["exclude"] + merged_config["exclude"],
                ignore_noqa=config_lint["ignore-noqa"],
                allowed_extensions=config_lint["allowed-extensions"],
                allowed_languages=config_lint["allowed-languages"],
                fix=config_lint["fix"],
                fixable=config_lint["fixable"],
                unfixable=config_lint["unfixable"],
                timestamp_column_suffix=config_lint["timestamp-column-suffix"],
                date_column_suffix=config_lint["date-column-suffix"],
                regex_partition=config_lint["regex-partition"],
                regex_index=config_lint["regex-index"],
                regex_constraint_primary_key=config_lint["regex-constraint-primary-key"],
                regex_constraint_unique_key=config_lint["regex-constraint-unique-key"],
                regex_constraint_foreign_key=config_lint["regex-constraint-foreign-key"],
                regex_constraint_check=config_lint["regex-constraint-check"],
                regex_constraint_exclusion=config_lint["regex-constraint-exclusion"],
                regex_sequence=config_lint["regex-sequence"],
                required_columns=[
                    Column(
                        name=column["name"],
                        data_type=column["data-type"],
                    )
                    for column in config_lint["required-columns"]
                ],
                disallowed_data_types=[
                    DisallowedDataType(
                        name=data_type["name"],
                        reason=data_type["reason"],
                        use_instead=data_type["use-instead"],
                    )
                    for data_type in config_lint["disallowed-data-types"]
                ],
                disallowed_schemas=[
                    DisallowedSchema(
                        name=schema["name"],
                        reason=schema["reason"],
                        use_instead=schema["use-instead"],
                    )
                    for schema in config_lint["disallowed-schemas"]
                ],
            ),
            format=Format(
                include=config_format["include"] + merged_config["include"],
                exclude=config_format["exclude"] + merged_config["exclude"],
                comma_at_beginning=config_format["comma-at-beginning"],
                new_line_before_semicolon=config_format["new-line-before-semicolon"],
                lines_between_statements=config_format["lines-between-statements"],
                remove_pg_catalog_from_functions=config_format[
                    "remove-pg-catalog-from-functions"
                ],
                diff=config_format["diff"],
                check=config_format["check"],
                no_cache=config_format["no-cache"],
            ),
        )
    except KeyError as error:
        msg = "Missing config key: "
        raise errors.MissingConfigError(msg + error.args[0]) from error



"""Core functionalities."""

from pgrubic.core import cache, config
from pgrubic.core.cache import Cache
from pgrubic.core.config import Config, parse_config
from pgrubic.core.linter import Linter, BaseChecker, ViolationStats
from pgrubic.core.loader import load_rules, load_formatters
from pgrubic.core.logger import logger
from pgrubic.core.filters import filter_sources
from pgrubic.core.formatter import Formatter

__all__ = [
    "BaseChecker",
    "Cache",
    "Config",
    "Formatter",
    "Linter",
    "ViolationStats",
    "add_apply_fix_to_rule",
    "add_set_locations_to_rule",
    "cache",
    "config",
    "filter_sources",
    "load_formatters",
    "load_rules",
    "logger",
    "parse_config",
]



"""Formatter for insert."""

from pglast import ast, enums, stream, printers


@printers.node_printer(ast.InferClause, override=True)
def infer_clause(node: ast.InferClause, output: stream.RawStream) -> None:
    """Printer for InferClause."""
    if node.conname:
        output.swrite("ON CONSTRAINT")
        output.space()
        output.print_name(node.conname)

    if node.indexElems:
        output.separator()
        with output.expression(need_parens=True):
            output.print_list(node.indexElems, standalone_items=False)

    if node.whereClause:
        output.newline()
        output.space()
        output.write("WHERE")
        output.space()
        output.print_node(node.whereClause)


@printers.node_printer(ast.OnConflictClause, override=True)
def on_conflict_clause(node: ast.OnConflictClause, output: stream.RawStream) -> None:
    """Printer for OnConflictClause."""
    on_conflict_action = enums.OnConflictAction
    if node.infer:
        output.print_node(node.infer)

    output.newline()

    if node.action == on_conflict_action.ONCONFLICT_NOTHING:
        output.space(4)
        output.write("DO NOTHING")
    elif node.action == on_conflict_action.ONCONFLICT_UPDATE:
        with output.push_indent(4):
            output.write("DO UPDATE SET")
            output.space()
            output.print_list(node.targetList)
            if node.whereClause:
                output.newline()
                output.space(4)
                output.write("WHERE")
                output.space()
                output.print_node(node.whereClause)


@printers.node_printer(ast.InsertStmt, override=True)
def insert_stmt(node: ast.InsertStmt, output: stream.RawStream) -> None:
    """Printer for InsertStmt."""
    with output.push_indent():
        if node.withClause:
            output.write("WITH")
            output.space()
            output.print_node(node.withClause)
            output.indent()

        output.write("INSERT INTO")
        output.space()
        output.print_node(node.relation)

        if node.cols:
            output.space()
            with output.expression(need_parens=True):
                output.print_list(node.cols, standalone_items=False)
        else:
            output.separator()

        if node.override:
            if node.override == enums.OverridingKind.OVERRIDING_USER_VALUE:
                output.space()
                output.write("OVERRIDING USER VALUE")
            elif node.override == enums.OverridingKind.OVERRIDING_SYSTEM_VALUE:
                output.space()
                output.write("OVERRIDING SYSTEM VALUE")

        if node.selectStmt:
            output.newline()
            output.print_node(node.selectStmt)
        else:
            output.write("DEFAULT VALUES")

        if node.onConflictClause:
            output.newline()
            output.space(4)
            output.write("ON CONFLICT")
            output.space() if node.onConflictClause.infer else output.write("")
            output.print_node(node.onConflictClause)

        if node.returningList:
            output.newline()
            output.write("RETURNING")
            output.space()
            output.print_name(node.returningList, ",")

        if node.withClause:
            output.dedent()



"""Test cache."""

import os
import typing
import pathlib
from unittest.mock import patch

from pgrubic import core

SOURCE_FILE: typing.Final[str] = "cache.sql"


def test_cache_file_existence(tmp_path: pathlib.Path, cache: core.Cache) -> None:
    """Test cache file existence."""
    source_code: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    source = directory / SOURCE_FILE
    source.write_text(source_code)

    assert not cache.cache_file.exists()

    cache.write(sources={source})

    assert cache.cache_file.exists()


def test_source_not_in_cache(tmp_path: pathlib.Path, cache: core.Cache) -> None:
    """Test source not in cache."""
    source_code: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    source = directory / SOURCE_FILE
    source.write_text(source_code)

    sources_to_be_formatted = cache.filter_sources(
        sources={source},
    )

    assert len(sources_to_be_formatted) == 1


def test_source_in_cache(tmp_path: pathlib.Path, cache: core.Cache) -> None:
    """Test source in cache."""
    source_code: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    source = directory / SOURCE_FILE
    source.write_text(source_code)

    sources_to_be_formatted = cache.filter_sources(
        sources={source},
    )

    assert len(sources_to_be_formatted) == 1

    cache.write(sources={source})

    sources_to_be_formatted = cache.filter_sources(
        sources={source},
    )

    assert len(sources_to_be_formatted) == 0


def test_source_invalidated_in_cache_by_modified_time(
    tmp_path: pathlib.Path,
    cache: core.Cache,
) -> None:
    """Test source invalidated in cache by modified time."""
    source_code: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    source = directory / SOURCE_FILE
    source.write_text(source_code)

    cache.write(sources={source})

    sources_to_be_formatted = cache.filter_sources(
        sources={source},
    )

    assert len(sources_to_be_formatted) == 0

    # Change the modified time of source
    os.utime(source, (1602179630, 1602179630))

    sources_to_be_formatted = cache.filter_sources(
        sources={source},
    )

    assert len(sources_to_be_formatted) == 1


def test_source_invalidated_in_cache_by_size(
    tmp_path: pathlib.Path,
    cache: core.Cache,
) -> None:
    """Test source invalidated in cache by size."""
    source_code: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    source = directory / SOURCE_FILE
    source.write_text(source_code)

    cache.write(sources={source})

    sources_to_be_formatted = cache.filter_sources(
        sources={source},
    )

    assert len(sources_to_be_formatted) == 0

    # Add a new line to source
    source.write_text(source_code + "\n")

    sources_to_be_formatted = cache.filter_sources(
        sources={source},
    )

    assert len(sources_to_be_formatted) == 1


def test_source_invalidated_in_cache_by_content(
    tmp_path: pathlib.Path,
    cache: core.Cache,
) -> None:
    """Test source invalidated in cache by content."""
    source_code: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    source = directory / SOURCE_FILE
    source.write_text(source_code)

    cache.write(sources={source})

    sources_to_be_formatted = cache.filter_sources(
        sources={source},
    )

    assert len(sources_to_be_formatted) == 0

    # Update a character in source
    source.write_text("SELECT b = NULL;")

    sources_to_be_formatted = cache.filter_sources(
        sources={source},
    )

    assert len(sources_to_be_formatted) == 1


def test_source_invalidated_in_cache_by_config(
    tmp_path: pathlib.Path,
    cache: core.Cache,
) -> None:
    """Test source invalidated in cache by config."""
    source_code: str = "SELECT a = NULL;"

    directory = tmp_path / "sub"
    directory.mkdir()

    source = directory / SOURCE_FILE
    source.write_text(source_code)

    cache.write(sources={source})

    sources_to_be_formatted = cache.filter_sources(
        sources={source},
    )

    assert len(sources_to_be_formatted) == 0

    cache.config.format.new_line_before_semicolon = True

    sources_to_be_formatted = cache.filter_sources(
        sources={source},
    )

    assert len(sources_to_be_formatted) == 1


def test_cache_directory_from_environment_variable_default_in_config(
    tmp_path: pathlib.Path,
) -> None:
    """Test cache directory from environment variable default in config."""
    config = core.parse_config()
    config.cache_dir = pathlib.Path(core.cache.DEFAULT_CACHE_DIR)

    with patch.dict(
        "os.environ",
        {core.cache.CACHE_DIR_ENVIRONMENT_VARIABLE: str(tmp_path)},
    ):
        cache = core.Cache(config=config)
        assert cache.config.cache_dir == tmp_path


def test_cache_directory_from_environment_variable_non_default_in_config(
    tmp_path: pathlib.Path,
) -> None:
    """Test cache directory from environment variable non default in config."""
    config = core.parse_config()
    config.cache_dir = pathlib.Path()

    with patch.dict(
        "os.environ",
        {core.cache.CACHE_DIR_ENVIRONMENT_VARIABLE: str(tmp_path)},
    ):
        cache = core.Cache(config=config)
        assert cache.config.cache_dir != tmp_path



"""Checker for insert without target columns."""

from pglast import ast, visitors

from pgrubic.core import linter


class InsertWithoutTargetColumns(linter.BaseChecker):
    """## **What it does**
    Checks for insert without target columns.

    ## **Why not?**
    Specifying the columns in a query explicitly greatly improves clarity and readability.
    This approach helps developers quickly grasp the purpose of a query and fosters
    better collaboration.

    Explicitly listing the target columns can help prevent errors

    ## **When should you?**
    Never.

    ## **Use instead:**
    Specify target columns.
    """

    def visit_InsertStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.InsertStmt,
    ) -> None:
        """Visit InsertStmt."""
        if node.selectStmt and not node.cols:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Insert statement without target columns",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Specify target columns",
                ),
            )



"""Non concurrent reindex."""

import typing

from pglast import ast, enums, visitors

from pgrubic.core import linter


class NonConcurrentReindex(linter.BaseChecker):
    """## **What it does**
    Checks non-concurrent reindex.

    ## **Why not?**
    Reindexing in non-concurrent mode will locks out writes (but not reads) on
    the table until it is done. This will cause downtime if the table is concurrently
    being written by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently written.

    ## **Use instead:**
    Reindex in concurrent mode: **REINDEX .. CONCURRENTLY ..**.
    """

    is_auto_fixable: bool = True

    def visit_ReindexStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ReindexStmt,
    ) -> None:
        """Visit ReindexStmt."""
        params: list[typing.Any] = (
            [param.defname for param in node.params] if node.params else []
        )

        if (
            node.kind != enums.ReindexObjectType.REINDEX_OBJECT_SYSTEM
            and "concurrently" not in params
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Non concurrent reindex",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Reindex in concurrent mode",
                ),
            )

            self._fix(node, params)

    def _fix(self, node: ast.ReindexStmt, params: list[typing.Any]) -> None:
        """Fix violation."""
        params.append(ast.DefElem(defname="concurrently"))

        node.params = params



"""Checker for SQL_ASCII encoding."""

from pglast import ast, visitors

from pgrubic.core import linter


class SqlAsciiEncoding(linter.BaseChecker):
    """## **What it does**
    Checks for SQL_ASCII encoding.

    ## **Why not?**
    SQL_ASCII means **no conversions** for the purpose of all encoding conversion
    functions. That is to say, the original bytes are simply treated as being in
    the new encoding, subject to validity checks, without any regard for what they mean.
    Unless extreme care is taken, an SQL_ASCII database will usually end up storing a
    mixture of many different encodings with no way to recover the original characters
    reliably.

    ## **When should you?**
    If your input data is already in a hopeless mixture of unlabelled encodings, such as
    IRC channel logs or non-MIME-compliant emails, then SQL_ASCII might be useful as a
    last resort—but consider using bytea first instead, or whether you could autodetect
    UTF8 and assume non-UTF8 data is in some specific encoding such as WIN1252.

    ## **Use instead:**
    UTF8
    """

    is_auto_fixable: bool = True

    def visit_DefElem(
        self,
        ancestors: visitors.Ancestor,
        node: ast.DefElem,
    ) -> None:
        """Visit DefElem."""
        if node.defname == "encoding" and node.arg.sval.lower() == "sql_ascii":
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="SQL_ASCII encoding detected",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use UTF8",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.DefElem) -> None:
        """Fix violation."""
        node.arg.sval = "utf8"



"""Checker for comparison with NULL."""

from pglast import ast, enums, visitors

from pgrubic import Operators
from pgrubic.core import linter


class NullComparison(linter.BaseChecker):
    """## **What it does**
    Checks for comparison with NULL.

    ## **Why not?**
    Comparing NULL to NULL with = returns NULL, not true.
    Comparing a value to NULL returns neither true nor false, but NULL.

    Do not write expression = NULL because NULL is not equal to NULL.
    (The null value represents an unknown value, and it is not known whether two unknown
    values are equal.)

    ## **When should you?**
    Never.

    ## **Use instead:**
    To check whether a value is or is not null, use the predicates:

    - expression IS NULL
    - expression IS NOT NULL
    """

    is_auto_fixable: bool = True

    def visit_A_Expr(
        self,
        ancestors: visitors.Ancestor,
        node: ast.A_Expr,
    ) -> ast.NullTest:
        """Visit A_Expr."""
        if (
            node.kind == enums.A_Expr_Kind.AEXPR_OP
            and node.name[-1].sval in (Operators.EQ, Operators.NOT_EQ)
            and any(
                isinstance(expr, ast.A_Const) and expr.isnull
                for expr in (node.rexpr, node.lexpr)
            )
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Comparison with NULL should be [IS | IS NOT] NULL",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use [IS | IS NOT] NULL",
                ),
            )

            return self._fix(node)

        return None  # pragma: no cover

    def _fix(self, node: ast.A_Expr) -> ast.NullTest:
        """Fix violation."""
        lexpr = (
            node.rexpr
            if not (isinstance(node.rexpr, ast.A_Const) and node.rexpr.isnull)
            else node.lexpr
        )

        null_type = (
            enums.NullTestType.IS_NULL
            if node.name[-1].sval == Operators.EQ
            else enums.NullTestType.IS_NOT_NULL
        )

        return ast.NullTest(arg=lexpr, nulltesttype=null_type)



"""Checker for constant generated columns."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class ConstantGeneratedColumn(linter.BaseChecker):
    """## **What it does**
    Checks for usage of constant generated columns.

    ## **Why not?**
    Always generating a constant value for a column is not useful and only leads to
    duplicated data all over the place and wastage of storage.

    ## **When should you?**
    If the table will be storing just one row at any point in time.

    ## **Use instead:**
    Generated column with expression.
    """

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> None:
        """Visit Constraint."""
        if node.contype == enums.ConstrType.CONSTR_GENERATED and isinstance(
            node.raw_expr,
            ast.A_Const,
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Generated column"
                    f" `{ancestors.find_nearest(ast.ColumnDef).node.colname}`"
                    " should not be a constant",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use an expression for the generated column",
                ),
            )



"""Checker for invalid sequence name according to naming convention."""

import re

from pglast import ast, enums, visitors

from pgrubic.core import linter


class InvalidSequenceName(linter.BaseChecker):
    """## **What it does**
    Checks that the name of the sequence to be created is valid according to naming
    convention.

    ## **Why not?**
    Naming conventions are crucial in database design as they offer consistency, clarity,
    and structure to the organization and accessibility of data within a database.

    A good naming convention makes your code easier to read and understand.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Name your sequence according to the set name convention.

    ## **Configuration**
    `regex-sequence`: Regex matching the naming convention for sequences.
    """

    def visit_CreateSeqStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateSeqStmt,
    ) -> None:
        """Visit CreateSeqStmt."""
        if not re.match(self.config.lint.regex_sequence, node.sequence.relname):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Sequence `{node.sequence.relname}` does not follow"
                    f" naming convention `{self.config.lint.regex_sequence}`",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Name your sequence according to the set naming convention",
                ),
            )

    def visit_RenameStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.RenameStmt,
    ) -> None:
        """Visit RenameStmt."""
        if node.renameType == enums.ObjectType.OBJECT_SEQUENCE and not re.match(
            self.config.lint.regex_sequence,
            node.newname,
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Sequence `{node.newname}` does not follow"
                    f" naming convention `{self.config.lint.regex_sequence}`",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Name your sequence according to the set naming convention",
                ),
            )



"""Checker for invalid check constraint name according to naming convention."""

import re

from pglast import ast, enums, visitors

from pgrubic.core import linter


class InvalidCheckConstraintName(linter.BaseChecker):
    """## **What it does**
    Checks that the name of the check constraint to be created is valid according to
    naming convention.

    ## **Why not?**
    Naming conventions are crucial in database design as they offer consistency, clarity,
    and structure to the organization and accessibility of data within a database.

    A good naming convention makes your code easier to read and understand.

    ## **When should you?**
    Never.

    ## **Use instead:**
    Name your check constraint according to the set name convention.

    ## **Configuration**
    `regex-constraint-check`: Regex matching the naming convention for check constraints.
    """

    def visit_Constraint(
        self,
        ancestors: visitors.Ancestor,
        node: ast.Constraint,
    ) -> None:
        """Visit Constraint."""
        if (
            node.contype == enums.ConstrType.CONSTR_CHECK
            and node.conname
            and (not re.match(self.config.lint.regex_constraint_check, node.conname))
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Check constraint"
                    f" `{node.conname}` does not follow naming convention"
                    f" `{self.config.lint.regex_constraint_check}`",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Name your check constraint according to the set naming"
                    " convention",
                ),
            )



"""Logger."""

import logging

from rich.logging import RichHandler

from pgrubic import PACKAGE_NAME

logging.basicConfig(
    datefmt="%Y-%m-%dT%H:%M:%S",
    format="%(message)s",
    handlers=[RichHandler(show_path=False)],
)

logger = logging.getLogger(PACKAGE_NAME)



"""Formatters for delete."""

from pglast import ast, stream, printers


@printers.node_printer(ast.DeleteStmt, override=True)
def delete_stmt(node: ast.DeleteStmt, output: stream.RawStream) -> None:
    """Printer for DeleteStmt."""
    with output.push_indent():
        if node.withClause:
            output.write("WITH")
            output.space()
            output.print_node(node.withClause)
            output.indent()

        output.write("DELETE FROM")
        output.space()
        output.print_node(node.relation)

        if node.usingClause:
            output.newline()
            output.space()
            output.write("USING")
            output.space()
            output.print_list(node.usingClause)

        if node.whereClause:
            output.newline()
            output.space()
            output.write("WHERE")
            output.space()
            output.print_node(node.whereClause)

        if node.returningList:
            output.newline()
            output.write("RETURNING")
            output.space()
            output.print_list(node.returningList, standalone_items=False)

        if node.withClause:
            output.dedent()



"""General rules."""

from pglast import ast

from pgrubic.core import config


def get_columns_from_table_creation(
    node: ast.CreateStmt,
) -> tuple[list[config.Column], set[str]]:
    """Get column details from table creation."""
    given_columns: list[config.Column] = []
    duplicate_columns: set[str] = set()

    if node.tableElts:
        given_columns = [
            config.Column(
                name=column.colname,
                data_type=column.typeName.names[-1],
            )
            for column in node.tableElts
            if isinstance(column, ast.ColumnDef)
        ]

        columns: list[str] = [column.name for column in given_columns]

        duplicate_columns = {column for column in columns if columns.count(column) > 1}

    return given_columns, duplicate_columns



"""Checker for usage of select into to create a new table."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class SelectInto(linter.BaseChecker):
    """## **What it does**
    Checks for usage of select into to create a new table.

    ## **Why not?**
    The SQL standard uses SELECT INTO to represent selecting values into scalar
    variables of a host program, rather than creating a new table.

    The PostgreSQL usage of SELECT INTO to represent table creation is historical.
    Some other SQL implementations also use SELECT INTO in this way (but most SQL
    implementations support CREATE TABLE AS instead). Apart from such compatibility
    considerations, it is best to use CREATE TABLE AS for this purpose in new code.

    ## **When should you?**
    Never.

    ## **Use instead:**
    CREATE TABLE AS.
    """

    is_auto_fixable: bool = True

    def visit_SelectStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.SelectStmt,
    ) -> ast.CreateTableAsStmt | None:
        """Visit SelectStmt."""
        if node.intoClause:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Use CREATE TABLE AS instead of SELECT INTO",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use CREATE TABLE AS",
                ),
            )

            return self._fix(node)

        return None

    def _fix(self, node: ast.SelectStmt) -> ast.CreateTableAsStmt:
        """Fix violation."""
        into_clause = node.intoClause

        # delete into clause
        node.intoClause = None

        return ast.CreateTableAsStmt(
            query=node,
            into=into_clause,
            objtype=enums.ObjectType.OBJECT_TABLE,
        )



"""Convention for extensions."""

from pglast import ast, visitors

from pgrubic.core import linter


class ProceduralLanguageWhitelist(linter.BaseChecker):
    """## **What it does**
    Checks that a procedural language to be created is allowed.

    ## **Why not?**
    By default, any procedural language can be loaded into the database.
    This is quite dangerous as some unsafe operations might be introuduced by languages.
    So you not only want to empower **CREATE LANGUAGE** to database owners, you
    also want to be able to review and explicitly allow procedural languages.

    ## **When should you?**
    Almost never. If a procedural language is not allowed, you are probably doing
    something wrong.

    ## **Use instead:**
    Procedural languages that are allowed.

    ## **Configuration**
    `allowed-languages`: List of allowed languages.
    """

    def visit_CreatePLangStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreatePLangStmt,
    ) -> None:
        """Visit CreatePLangStmt."""
        if (
            node.plname not in self.config.lint.allowed_languages
            and "*" not in self.config.lint.allowed_languages
        ):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description=f"Language '{node.plname}' is not allowed",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="If you want to allow this language, add the language to the"
                    " allowed languages list",
                ),
            )



"""Checker for usage of AStar."""

from pglast import ast, visitors

from pgrubic.core import linter


class Asterisk(linter.BaseChecker):
    """## **What it does**
    Checks for usage of asterisk (*) in column references.

    ## **Why not?**
    Specifying the columns in a query explicitly greatly improves clarity and readability.
    This approach helps developers quickly grasp the purpose of the query and fosters
    better collaboration.

    Also, using (SELECT *) complicates code maintenance. When the table structure
    changes, such as adding, renaming, or removing columns, queries with SELECT * can
    fail unexpectedly or silently return incorrect results.
    By explicitly listing the necessary columns, you ensure the code is more resilient to
    changes in the database schema.

    ## **When should you?**
    Almost Never.

    ## **Use instead:**
    Name Columns Explicitly.
    """

    def visit_A_Star(
        self,
        ancestors: visitors.Ancestor,
        node: ast.A_Star,
    ) -> None:
        """Visit A_Star."""
        if ancestors.find_nearest(ast.SelectStmt):
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Asterisk in column reference is discouraged",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Name columns explicitly",
                ),
            )



"""Filters."""

import fnmatch
import pathlib


def filter_sources(
    *,
    sources: tuple[pathlib.Path, ...],
    include: list[str],
    exclude: list[str],
    respect_gitignore: bool,
    extension: str = "sql",
) -> set[pathlib.Path]:
    """Filter sources base on include and exclude and either respect gitignore.

    Paramaters:
    -----------
    sources: tuple[pathlib.Path, ...]
        List of sources to filter.
    include: list[str]
        List of file patterns to include.
    exclude: list[str]
        List of file patterns to exclude.
    respect_gitignore: bool
        Whether to respect gitignore.
    extension: str
        File extension to filter. Default is "sql".

    Returns:
    -------
    set[pathlib.Path]
        Set of filtered sources.

    """
    flattened_sources: set[pathlib.Path] = set()

    for source in sources:
        if source.is_dir():
            flattened_sources.update(source.glob(f"**/*.{extension}"))

        elif source.suffix == f".{extension}":
            flattened_sources.add(source)

    included_sources: set[pathlib.Path] = set()

    for source in flattened_sources:
        if (
            _is_file_included(
                source=str(source),
                include=include,
                exclude=exclude,
            )
            and source.stat().st_size > 0
        ):
            if respect_gitignore and _is_git_ignored(str(source)):
                continue

            included_sources.add(source)

    return included_sources


def _is_file_included(
    *,
    source: str,
    include: list[str],
    exclude: list[str],
) -> bool:
    """Check if a source should be included or excluded based on global config.

    Paramaters:
    -----------
    source: str
        Path to the source file.
    include: list[str]
        List of file patterns to include.
    exclude: list[str]
        List of file patterns to exclude.

    Returns:
    -------
    bool
        True if the source should be included, False otherwise.

    """
    return bool(
        (not include or any(fnmatch.fnmatch(source, pattern) for pattern in include))
        and not any(fnmatch.fnmatch(source, pattern) for pattern in exclude),
    )


def _is_git_ignored(source: str) -> bool:
    """Check if a source is git ignored.

    Paramaters:
    -----------
    source: str
        Path to the source file.

    Returns:
    -------
    bool
        True if the source is git ignored, False otherwise.

    """
    try:
        import git
        import git.exc

        repo = git.Repo(source, search_parent_directories=True)
        return bool(repo.ignored(source))
    except ImportError:  # pragma: no cover
        # git is not installed
        return False
    except (git.exc.InvalidGitRepositoryError, git.exc.GitCommandError):
        return False



"""Checker for ID column."""

from pglast import ast, enums, visitors

from pgrubic.core import linter


class IdColumn(linter.BaseChecker):
    """## **What it does**
    Checks for usage of ID columns.

    ## **Why not?**
    The name "id" does not provide much information about what the column represents in
    the context of the table as it is so generic. Using a more descriptive name can
    improve clarity, readability, and maintainability of the database schema.

    ## **When should you?**
    Almost never.

    ## **Use instead:**
    Descriptive name.
    """

    is_auto_fixable: bool = True

    def visit_ColumnDef(
        self,
        ancestors: visitors.Ancestor,
        node: ast.ColumnDef,
    ) -> None:
        """Visit ColumnDef."""
        if (
            (
                ancestors.find_nearest(ast.AlterTableCmd)
                and ancestors.find_nearest(ast.AlterTableCmd).node.subtype
                == enums.AlterTableType.AT_AddColumn
            )
            or ancestors.find_nearest(ast.CreateStmt)
        ) and node.colname.lower() == "id":
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Use descriptive name for column instead of"
                    f" `{node.colname}`",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use a more descriptive name",
                ),
            )

            self._fix(ancestors, node)

    def _fix(self, ancestors: visitors.Ancestor, node: ast.ColumnDef) -> None:
        """Fix violation."""
        if ancestors.find_nearest(ast.AlterTableCmd):
            table = ancestors.parent.parent.node.relation.relname

        if ancestors.find_nearest(ast.CreateStmt):
            table = ancestors.parent.node.relation.relname

        node.colname = table + "_" + node.colname



"""Formatter for CTEs."""

from pglast import ast, stream, printers


@printers.node_printer(ast.WithClause, override=True)
def with_clause(node: ast.WithClause, output: stream.RawStream) -> None:
    """Printer for WithClause."""
    relative_indent = -2

    if node.recursive:
        relative_indent -= output.write("RECURSIVE ")

    output.print_list(node.ctes, relative_indent=relative_indent, standalone_items=False)


@printers.node_printer(ast.CommonTableExpr, override=True)
def common_table_expr(node: ast.CommonTableExpr, output: stream.RawStream) -> None:
    """Printer for CommonTableExpr."""
    output.print_name(node.ctename)
    if node.aliascolnames:
        output.space()
        with output.expression(need_parens=True):
            output.print_name(node.aliascolnames, ",")
        output.indent(amount=-1, relative=False)

    output.swrite("AS")
    printers.dml.cte_materialize_printer(node.ctematerialized, node, output)
    output.space()

    with output.expression(need_parens=False):
        output.write("(")
        output.newline()
        output.indent(4)
        output.print_node(node.ctequery)
        output.indent(amount=-4, relative=False)
        output.newline()
        output.write(")")

    if node.search_clause:
        output.space()
        output.print_node(node.search_clause)

    if node.cycle_clause:
        output.space()
        output.print_node(node.cycle_clause)
    if node.aliascolnames:
        output.dedent()
    output.newline()



"""Checker for inline sql function body with wrong language."""

from pglast import ast, visitors

from pgrubic.core import linter


class InlineSqlFunctionBodyWrongLanguage(linter.BaseChecker):
    """## **What it does**
    Checks for inline sql function bodies which are not defined with LANGUAGE SQL.

    ## **Why not?**
    Inline SQL function bodies are only valid for language SQL.
    While PostgreSQL parses such functions, they fail at creation with the error below:

    ```bash
    ERROR:  inline SQL function body only valid for language SQL
    ```

    ## **When should you?**
    Never.

    ## **Use instead:**
    LANGUAGE SQL.
    """

    is_auto_fixable: bool = True

    def visit_CreateFunctionStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.CreateFunctionStmt,
    ) -> None:
        """Visit CreateFunctionStmt."""
        language: str | None = next(
            (
                option.arg.sval
                for option in node.options
                if option.defname.upper() == "LANGUAGE"
            ),
            None,
        )

        if node.sql_body and language and language.upper() != "SQL":
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Inline SQL function body is only valid for language SQL",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Use LANGUAGE SQL",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.CreateFunctionStmt) -> None:
        """Fix violation."""
        for option in node.options:
            if option.defname.upper() == "LANGUAGE":
                option.arg.sval = "SQL"



"""Non-concurrent index creation."""

from pglast import ast, visitors

from pgrubic.core import linter


class NonConcurrentIndexCreation(linter.BaseChecker):
    """## **What it does**
    Checks non-concurrent index creation.

    ## **Why not?**
    Creating an index in non-concurrent mode will locks out writes (but not reads) on the
    table until it is done. This will cause downtime if the table is concurrently being
    written by other clients.

    ## **When should you?**
    If the table is empty.
    If the table is not empty but is not being concurrently written.

    ## **Use instead:**
    Create the index in concurrent mode: **CREATE .. INDEX CONCURRENTLY ON ..**.
    """

    is_auto_fixable: bool = True

    def visit_IndexStmt(
        self,
        ancestors: visitors.Ancestor,
        node: ast.IndexStmt,
    ) -> None:
        """Visit IndexStmt."""
        if not node.concurrent:
            self.violations.add(
                linter.Violation(
                    rule_code=self.code,
                    rule_name=self.name,
                    rule_category=self.category,
                    line_number=self.line_number,
                    column_offset=self.column_offset,
                    line=self.line,
                    statement_location=self.statement_location,
                    description="Non concurrent index creation",
                    is_auto_fixable=self.is_auto_fixable,
                    is_fix_enabled=self.is_fix_enabled,
                    help="Create the index in concurrent mode",
                ),
            )

            self._fix(node)

    def _fix(self, node: ast.IndexStmt) -> None:
        """Fix violation."""
        node.concurrent = True



"""Formatter for table."""

from pglast import ast, enums, stream, printers

from pgrubic.formatters.ddl import IF_EXISTS, IF_NOT_EXISTS


@printers.node_printer(ast.IntoClause, override=True)
def into_clause(node: ast.IntoClause, output: stream.RawStream) -> None:
    """Printer for IntoClause."""
    output.print_node(node.rel)

    if node.colNames:
        output.space()
        with output.expression(need_parens=True):
            output.print_name(node.colNames, ",")

    if node.accessMethod:
        output.newline()
        output.space()
        output.writes("USING")
        output.print_name(node.accessMethod)

    if node.options:
        output.newline()
        output.space(2)
        output.write("WITH")
        output.space()
        with output.expression(need_parens=True):
            output.newline()
            output.space(6)
            output.print_list(node.options)
            output.newline()
            output.space(2)

    if node.onCommit != enums.OnCommitAction.ONCOMMIT_NOOP:
        output.space()
        output.write("ON COMMIT")
        output.space()
        if node.onCommit == enums.OnCommitAction.ONCOMMIT_PRESERVE_ROWS:
            output.write("PRESERVE ROWS")
        elif node.onCommit == enums.OnCommitAction.ONCOMMIT_DELETE_ROWS:
            output.write("DELETE ROWS")
        elif node.onCommit == enums.OnCommitAction.ONCOMMIT_DROP:
            output.write("DROP")

    if node.tableSpaceName:
        output.newline()
        output.write("TABLESPACE")
        output.space()
        output.print_name(node.tableSpaceName)


@printers.node_printer(ast.PartitionSpec, override=True)
def partition_spec(node: ast.PartitionSpec, output: stream.RawStream) -> None:
    """Printer for PartitionSpec."""
    strategy = {
        enums.PartitionStrategy.PARTITION_STRATEGY_LIST: "LIST",
        enums.PartitionStrategy.PARTITION_STRATEGY_RANGE: "RANGE",
        enums.PartitionStrategy.PARTITION_STRATEGY_HASH: "HASH",
    }[node.strategy]

    output.print_symbol(strategy)
    output.space()
    with output.expression(need_parens=True):
        output.print_list(nodes=node.partParams, standalone_items=False)


@printers.node_printer(ast.CreateTableAsStmt, override=True)
def create_table_as_stmt(node: ast.CreateTableAsStmt, output: stream.RawStream) -> None:
    """Printer for CreateTableAsStmt."""
    output.writes("CREATE")
    output.space()

    if node.into.rel.relpersistence == enums.RELPERSISTENCE_TEMP:
        output.writes("TEMPORARY")
    elif node.into.rel.relpersistence == enums.RELPERSISTENCE_UNLOGGED:
        output.writes("UNLOGGED")

    output.writes(printers.ddl.OBJECT_NAMES[node.objtype])

    if node.if_not_exists:
        output.writes(IF_NOT_EXISTS)

    output.print_node(node.into)
    output.swrite("AS")
    output.newline()
    output.print_node(node.query)

    if node.into.skipData:
        output.newline()
        output.space(2)
        output.write("WITH NO DATA")


@printers.node_printer(ast.CreateForeignTableStmt, override=True)
def create_foreign_table_stmt(
    node: ast.CreateForeignTableStmt,
    output: stream.RawStream,
) -> None:
    """Printer for CreateForeignTableStmt."""
    output.print_node(node.base)
    output.newline()

    if node.base.partbound:
        output.space(4)

    output.write("SERVER")
    output.space()
    output.print_name(node.servername)

    if node.options:
        output.newline()
        if node.base.partbound:
            output.space(4)
        with output.push_indent():
            output.write("OPTIONS")
            output.space()
            with output.expression(need_parens=True):
                output.newline()
                output.space(4)
                output.print_list(node.options)
                output.newline()


@printers.node_printer(ast.CreateStmt, override=True)
def create_stmt(
    node: ast.CreateStmt,
    output: stream.RawStream,
) -> None:
    """Printer for CreateStmt."""
    output.writes("CREATE")

    if isinstance(node.ancestors[0], ast.CreateForeignTableStmt):
        output.writes("FOREIGN")
    elif node.relation.relpersistence == enums.RELPERSISTENCE_TEMP:
        output.writes("TEMPORARY")
    elif node.relation.relpersistence == enums.RELPERSISTENCE_UNLOGGED:
        output.writes("UNLOGGED")

    output.writes("TABLE")

    if node.if_not_exists:
        output.writes(IF_NOT_EXISTS)

    output.print_node(node.relation)

    if node.ofTypename:
        output.space()
        output.write("OF")
        output.space()
        output.print_name(node.ofTypename)

    if node.partbound:
        output.newline()
        output.space(4)
        output.write("PARTITION OF")
        output.space()
        output.print_list(node.inhRelations)

    if node.tableElts:
        # move table constraints to the end
        columns = [x for x in node.tableElts if not isinstance(x, ast.Constraint)] + [
            x for x in node.tableElts if isinstance(x, ast.Constraint)
        ]
        output.space()
        with output.expression(need_parens=True):
            output.newline()
            output.space(4)
            output.print_list(columns)
            output.newline()
    elif node.partbound:
        output.write("")
    elif not node.ofTypename:
        output.space()
        output.write("()")

    if node.inhRelations and not node.partbound:
        output.newline()
        output.write("INHERITS")
        output.space()
        with output.expression(need_parens=True):
            output.print_list(node.inhRelations)

    if node.partbound:
        output.newline()
        output.space(4)
        output.print_node(node.partbound)

    if node.partspec:
        output.newline()
        output.writes("PARTITION BY")
        output.print_node(node=node.partspec)

    if node.oncommit != enums.OnCommitAction.ONCOMMIT_NOOP:
        output.newline()
        output.write("ON COMMIT")
        output.space()
        if node.oncommit == enums.OnCommitAction.ONCOMMIT_PRESERVE_ROWS:
            output.write("PRESERVE ROWS")
        elif node.oncommit == enums.OnCommitAction.ONCOMMIT_DELETE_ROWS:
            output.write("DELETE ROWS")
        elif node.oncommit == enums.OnCommitAction.ONCOMMIT_DROP:
            output.write("DROP")

    if node.accessMethod:
        output.newline()
        output.space()
        output.writes("USING")
        output.print_name(node.accessMethod)

    if node.options:
        output.newline()
        output.space(2)
        output.write("WITH")
        output.space()
        with output.expression(need_parens=True):
            output.newline()
            output.space(6)
            output.print_list(node.options)
            output.newline()
            output.space(2)

    if node.tablespacename:
        output.newline()
        output.write("TABLESPACE")
        output.space()
        output.print_name(node.tablespacename)


@printers.node_printer(ast.AlterTableStmt, override=True)
def alter_table_stmt(node: ast.AlterTableStmt, output: stream.RawStream) -> None:
    """Printer for AlterTableStmt."""
    output.write("ALTER")
    output.space()
    output.writes(printers.ddl.OBJECT_NAMES[node.objtype])

    if node.missing_ok:
        output.write(IF_EXISTS)

    output.space()
    output.print_node(node.relation)
    output.newline()
    output.space(4)
    output.print_list(nodes=node.cmds, standalone_items=True)
